#ifndef OOO_INSTRUMENTATION_H_INCLUDED_
#define OOO_INSTRUMENTATION_H_INCLUDED_

#include <iostream>
#include <cstdlib>

#include "rose.h"
#include "../segment_graph/combined_data_conflict_table.h"
#include "../segment_graph/prediction_time_advance_table.h"
#include "../segment_graph/prediction_event_notification_table.h"

namespace risc {

class Design;

namespace sg {

class DataConflictTable;
class EventConflictTable;
class SegmentGraph;
class TimeAdvanceTable;
}

namespace tools {

class ParseStatus;
}

namespace inst {

/**
 * \class InstrumentationTraversal
 * \brief The AST traversal class
 */
class InstrumentationTraversal: public AstSimpleProcessing
{
  public:
    InstrumentationTraversal(SgType*, SgExpression*, SgExpression*,
                             SgClassDefinition*, std::map<SgNode*, int>&);
    virtual void visit(SgNode* astNode);

  private:
    /**
     * \brief Type of ::sc_core::chnl_scoped_lock
     */ 
    SgType* variableType;

    /**
     * \brief Reference expression of m_mutex in ::sc_core::sc_prim_channel  
     */  
    SgExpression* PrimChnlVarRefExp;

    /**
     * \brief Reference expression of m_mutex in ::sc_core::sc_channel 
     */
    SgExpression* SCChnlVarRefExp;

    /**
     * \brief SgClassDefinition of "::sc_core::sc_module"
     */ 
    SgClassDefinition* SCModuleClassDef;

    /**
     * \brief Mapping from SgNode to segment ID
     */  
    std::map<SgNode*, int>& FuncCallExptoID;
};

/**
 * \fn risc_link
 * \brief Tailored SgProject::link for customized linker. The original ROSE
 *        codes are at ~src/frontend/SageIII/sage_support/sage_support.cpp
 * \return The status code from the vendor linker
 */
int
risc_link(SgProject* project, risc::tools::ParseStatus* parseStatus, const std::string tool_name);

/**
 * \fn risc_buildCompilerCommandLineOptions
 * \brief Tailored SgFile::buildCompilerCommandLineOptions for customized 
 *        compiler. The original ROSE codes are at
 *        ~src/frontend/SageIII/sage_support/cmdline.cpp
 */
void
risc_buildCompilerCommandLineOptions(SgFile* file,
                                     std::vector<std::string>& argv,
                                     std::vector<std::string>& compilerCmdLine,
                                     risc::tools::ParseStatus* parseStatus,
                                     const std::string tool_name);

/**
 * \fn risc_compileOutputFile
 * \brief Tailored SgFile::compileOutput for customized compiler.
 *        The original ROSE codes are at ~src/ROSETTA/Grammer/Support.code and 
 *        ~src/frontend/SageIII/sage_support/sage_support.cpp
 * \return The status code from the vendor compiler 
 */
int
risc_compileOutputFile(SgFile* file, risc::tools::ParseStatus* parseStatus, const std::string tool_name);

/**
 * \fn risc_compileOutput
 * \brief Tailored SgProject::compileOutput for customized compiler.
 *        The original ROSE codes are at
 *        ~src/frontend/SageIII/sage_support/sage_support.cpp
 * \return The status code from the vendor compiler 
 */
int
risc_compileOutput(SgProject* project, risc::tools::ParseStatus* parseStatus, const std::string tool_name);

/**
 * \fn risc_unparse
 * \brief This function unparses AST trees to source code file(s)
 */
void
risc_unparse(risc::Design &design,
    risc::tools::ParseStatus* parseStatus,
    const std::string tool_name,
    UnparseFormatHelp *unparseFormatHelp = NULL,
    UnparseDelegate* unparseDelegate = NULL);

/**
 * \fn risc_patch
 * \brief This function patches the static_cast issues in the unparsed
 *        source code file(s)
 */
void
risc_patch(risc::Design &design,
    risc::tools::ParseStatus* parseStatus,
    const char *methodPattern,
    const char *threadPattern,
    const char *cthreadPattern,
    const std::string tool_name);

/**
 * \fn risc_compile
 * \brief This function compiles and links the unparsed source code file(s)
 *        using the customized compiler
 * \return  The status code from the vendor compiler
 */
int
risc_compile(risc::Design &design, risc::tools::ParseStatus* parseStatus,
    const std::string tool_name);

/**
 * \fn risc_backend
 * \brief Tailored backend for SystemC source files. The original ROSE codes 
 *        are at ~src/roseSupport/utility_functions.C
 * \return The status code from the vendor compiler 
 */
int
risc_backend(risc::Design &design,
    risc::tools::ParseStatus* parseStatus,
    const std::string tool_name,
    UnparseFormatHelp *unparseFormatHelp = NULL,
    UnparseDelegate* unparseDelegate = NULL);

/**
 * \fn instrumentor
 * \brief This function instruments the AST tree of SystemC designs, and 
 *        compiles unparsed source file(s) using customized vendor compiler
 */
void
instrumentor(risc::Design &design,
    risc::sg::SegmentGraph &sg,
    risc::sg::DataConflictTable* data_conflict_table,
    risc::sg::EventConflictTable* event_conflict_table,
    risc::sg::TimeAdvanceTable* time_advance_table,
    risc::sg::CombinedDataConflictTable* combined_data_conflict_table,
    risc::sg::PredictionTimeAdvanceTable* prediction_time_advance_table,
    risc::sg::PredictionEventNotificationTable* prediction_event_notification_table,
    risc::tools::ParseStatus* parseStatus);


/**
 * \fn    instrument_port_calls
 * \brief instruments the upcoming segments ids to thread local data
 */
void instrument_port_calls(risc::sg::SegmentGraph &sg);

/**
 * \fn    instrument_wait_in_channels
 * \brief changes the segment id passing from constants to variables in channels
 *        which are passed by thread local data
 */
void instrument_wait_in_channels(risc::sg::SegmentGraph &sg);

/**
 * \fn    generate_unique_name
 * \brief generates a unique variable based on var_name
 */
std::string generate_unique_name(std::string var_name);

/**
 * \brief counter to get a unique variable name with number
 */
extern int counter;

} // end of namespace inst

} // end of namespace risc

#endif /* OOO_INSTRUMENTATION_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
