#include "design.h"

#include <vector>
#include <ctime>
#include <string.h>

#include "variable.h"

#include "cthread.h"
#include "dynamic_internal_representation.h"
#include "dynamic_path_instance_mapper.h"
#include "event.h"
#include "function.h"
#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "inoutport.h"
#include "inport.h"
#include "instance.h"
#include "instance_tree.h"
#include "interface.h"
#include "method.h"
#include "module.h"
#include "module_instance.h"
#include "outport.h"
#include "primitive_channel.h"
#include "static_path_instance_mapper.h"
#include "thread.h"
#include "../ast_traverser/traverser.h"
#include "../ast_traverser/path_checker.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/command_line_parsing.h"
#include "../tools/file_classifier.h"
#include "../tools/function_helper.h"
#include "../tools/misc.h"
#include "../tools/rose_nodes.h"
#include "../tools/systemc_definitions.h"
#include "../instrumentation/dyn_analysis_instr.h"

void yyerror(const char *s);
int yylex();

risc::Design::Design(int argc, char* argv[])
{
  risc::GlobalDesign::design_static = this;

  old_argc_ = argc;
  old_argv_ = argv;

  // here we are parsing out the filename for the dynamic analysis
  dynamic_analysis_filename_
    = tools::dynamic_analysis_input_filename(argc, argv);

  instrumentor_output_filename_
    = tools::instrumentor_output_filename(argc, argv);

  project_ = frontend(argc, argv);

  if(project_ == NULL) {
    std::cout << std::endl << std::endl;
    std::cout << "Error: The AST for this design could not be created."
              << std::endl
              << "No further analysis possible."
              << std::endl;
    exit(1);
  }

  // the ordering is important
  initialize_interfaces();
  initialize_definitions();
  initialize_top_modules();
  initialize_global_variables();
  initialize_global_functions();
  initialize_path_instance_mapper();
  initialize_port_mapping();

  // check if the design is analyzable for segment graph generation
  if(!is_analyzable()) {
    std::cout << std::endl << std::endl;
    std::cout << "Error: This design is not analyzable."
              << std::endl
              << "At least one module/channel has a port which is mapped to "
              << "two or more channels with different interfaces."
              << std::endl;

    exit(1);
  }
}

risc::Design::~Design()
{
  // clearing all vectors
  for(ModuleVectorIter iter = module_definitions_.begin();
      iter != module_definitions_.end(); iter++) {
    delete *iter;
  }

  for(PrimitiveChannelVectorIter
      iter  = primitive_channel_.begin();
      iter != primitive_channel_.end(); iter++) {
    delete *iter;
  }

  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end(); iter++) {
    delete *iter;
  }

  delete path_instance_mapper_;
}

void risc::Design::initialize_definitions()
{
  std::vector<SgClassDefinition*> channel
    = risc::traverser::find_all_channels();

  for(std::vector<SgClassDefinition*>::iterator
      iter  = channel.begin();
      iter != channel.end();
      iter++) {

    SgClassDefinition* channel = isSgClassDefinition(*iter);
    bool is_already_inserted = false;

    for(HierarchicalChannelVectorIter
        channel_iter  = hierarchical_channel_.begin();
        channel_iter != hierarchical_channel_.end();
        channel_iter++) {

      if((*channel_iter)->get_ast_node() == channel) {
        is_already_inserted = true;
      }
    }

    if(is_already_inserted) {
      continue;
    }

    if(tools::is_derived_from_sc_channel(channel)) {

      hierarchical_channel_.push_back(new HierarchicalChannel(channel));
    }

    for(PrimitiveChannelVectorIter
        channel_iter  = primitive_channel_.begin();
        channel_iter != primitive_channel_.end();
        channel_iter++) {

      if((*channel_iter)->get_ast_node() == channel) {
        is_already_inserted = true;
      }
    }

    if(is_already_inserted) {
      continue;
    }

    if(tools::is_derived_from_sc_prim_channel(channel)) {

      primitive_channel_.push_back(new PrimitiveChannel(channel));
    }
  }

  // Search for all class definitions that are derived from sc_module and
  // are not defined in a file in the include path
  std::vector<SgClassDefinition*> modules = risc::traverser::find_all_modules();

  for(std::vector<SgClassDefinition*>::iterator
      iter  = modules.begin();
      iter != modules.end();
      iter++) {

    SgClassDefinition* temp_cd = isSgClassDefinition(*iter);
    bool is_already_inserted = false;

    for(ModuleVectorIter
        module_iter  = module_definitions_.begin();
        module_iter != module_definitions_.end();
        module_iter++) {

      if((*module_iter)->get_ast_node() == temp_cd) {
        is_already_inserted = true;
      }
    }

    if(is_already_inserted) {
      continue;
    }

    if(temp_cd) {

      module_definitions_.push_back(new Module(temp_cd));
    }
  }
}

risc::Module* risc::Design::find_module_definition(SgClassDefinition* cd)
{
  Module* temp = NULL;

  for(ModuleVectorIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {

    if(risc::tools::class_types_are_identical((*iter)->get_ast_node(),cd)) {
      temp = *iter;
    }
  }

  return temp;
}

risc::HierarchicalChannel*
risc::Design::find_channel_definition(SgClassDefinition* cd)
{
  HierarchicalChannel* temp = NULL;

  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin() ;
      iter != hierarchical_channel_.end();
      iter++) {

    if(risc::tools::class_types_are_identical((*iter)->get_ast_node(),cd)) {
      temp = *iter;
    }
  }

  return temp;
}

risc::PrimitiveChannel*
risc::Design::find_prim_channel_definition(SgClassDefinition* cd)
{
  PrimitiveChannel* temp = NULL;

  for(PrimitiveChannelVectorIter
      iter  = primitive_channel_.begin() ;
      iter != primitive_channel_.end();
      iter++) {

    if(risc::tools::class_types_are_identical((*iter)->get_ast_node(),cd)) {
      temp = *iter;
    }
  }

  return temp;
}

risc::Interface*
risc::Design::find_interface(SgClassDefinition* cd)
{
  Interface* temp = NULL;

  for(InterfaceVectorIter
      iter  = interfaces_.begin() ;
      iter != interfaces_.end();
      iter++) {

    if(risc::tools::class_types_are_identical((*iter)->get_ast_node(),cd)) {
      temp = *iter;
    }
  }

  return temp;
}

void risc::Design::initialize_top_modules()
{
  //FIXME get sc_main function and search for members that are variable
  //      instances and derived from sc_module

  SgFunctionDefinition* func_def
    = tools::SystemCDefinitions::get_sc_main();
  SgBasicBlock* bb = func_def->get_body();

  for(std::vector<SgStatement*>::iterator
      iter  = bb->get_statements().begin();
      iter != bb->get_statements().end();
      iter++) {

    SgVariableDeclaration* var_decl =  isSgVariableDeclaration(*iter);

    if(var_decl) {

      SgType* type = var_decl->get_decl_item(0)->get_type();

      // in this case we are addressing an int, float, ...
      if(tools::is_builtin_type(type))
        continue;

      SgArrayType* array_type = isSgArrayType(type);

      // in this case we are addressing a array of int, float, ...
      if(array_type)
        if(tools::is_builtin_type(SageInterface::getArrayElementType(type)))
          continue;

      // At this point we can know that it is a class or a corresponding array
      SgClassDefinition* class_def = NULL;

      // If it is an array we have to get the definition.
      // So, we are asking for the definition behind the array
      if(array_type) {

        SgClassType* class_type =
          isSgClassType(SageInterface::getArrayElementType(type));


        SgClassDeclaration* class_decl =
          isSgClassDeclaration(class_type->get_declaration());

        class_decl
          = isSgClassDeclaration(class_decl->get_definingDeclaration());

        class_def = class_decl->get_definition();

      } else {
        class_def = tools::class_definition_of_variable(var_decl);
      }

      if(tools::is_derived_from_sc_channel(class_def)) {

        HierarchicalChannel* hc = find_channel_definition(class_def);

        top_modules_.push_back(
                         new HierarchicalChannelInstance(
                                 var_decl->get_definition(), hc));
        continue;
      }

      if(tools::is_derived_from_sc_module(class_def)) {

        Module* md = find_module_definition(class_def);

        top_modules_.push_back(
                         new ModuleInstance(var_decl->get_definition(), md));
      }
    }
  }
}

void risc::Design::initialize_global_variables()
{
  SgDeclarationStatementPtrList &global_decls
    = SageInterface::getGlobalScope(
        tools::SystemCDefinitions::get_sc_main())->get_declarations();

  int counter = 0;

  for(SgDeclarationStatementPtrList::iterator
      iter  = global_decls.begin();
      iter != global_decls.end();
      iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);

    if(var_decl) {

      // check if we added the variable already
      bool exists_already = false;
      for(VariableVectorIter
          iter  = global_variables_.begin();
          iter != global_variables_.end();
          iter++) {

        if(var_decl->get_decl_item(0)->search_for_symbol_from_symbol_table()
            == (*iter)->get_ast_node()->get_declaration()->
                   get_decl_item(0)->search_for_symbol_from_symbol_table()) {
          exists_already = true;
        }
      }

      if(exists_already) {
        continue;
      }

      // here we are checking the variable is module
      SgClassType *class_type
        = isSgClassType(var_decl->get_decl_item(0)->get_type());

      // in the following we are checking if the global variable is a potential
      // top module
      if(class_type) {
        SgClassDeclaration *class_decl
          = isSgClassDeclaration(class_type->get_declaration());
        class_decl
          = isSgClassDeclaration(class_decl->get_definingDeclaration());

        if(class_decl) {
          SgClassDefinition *class_def = class_decl->get_definition();

          if(class_def) {

            if(tools::is_derived_from_sc_channel(class_def)) {

              Module* md = find_module_definition(class_def);

              top_modules_.
                push_back(
                    new ModuleInstance(var_decl->get_definition(), md));
              continue;
            }

            if(tools::is_derived_from_sc_module(class_def)) {

              Module* md = find_module_definition(class_def);

              top_modules_.
                push_back(
                    new ModuleInstance(var_decl->get_definition(), md));
            }
          }
        }
      }

      global_variables_.push_back(new Variable(var_decl->get_definition()));
    }
  }
}

void risc::Design::initialize_global_functions()
{
  std::vector<SgNode*> function_definition
    = tools::RoseNodes::get_func_defs();

  int filter = 0;
  for(std::vector<SgNode*>::const_iterator
      i  = function_definition.begin();
      i != function_definition.end(); ++i) {

    SgFunctionDefinition *func = isSgFunctionDefinition(*i);

    // Theses functions are not in the global scope
    if(SageInterface::getScope(func->get_declaration()) !=
        SageInterface::getGlobalScope(func->get_declaration())) {
      continue;
    }

    // Skip all template functions
    if(isSgTemplateFunctionDefinition(func) ||
        isSgTemplateInstantiationFunctionDecl(func->get_declaration()) ||
        isSgTemplateInstantiationMemberFunctionDecl(func->get_declaration())) {
      continue;
    }

    if(func->get_file_info()->isCompilerGenerated()) {
      continue;
    }

    // Skip all functions which are part of a class
    if(func->get_declaration()->get_specialFunctionModifier().isConstructor() ||
        func->get_declaration()->get_specialFunctionModifier().isDestructor() ||
        isSgMemberFunctionDeclaration(func->get_declaration())) {
      continue;
    }

    // These functions are built in functions. So, we can skip them
    if(risc::tools::is_builtin_function(func->get_declaration())) {
      continue;
    }

    // Global functions are not allowed to start with a '_'
    // https://gcc.gnu.org/onlinedocs/libstdc++/manual/source_code_style.html
    if(SageInterface::get_name(func).at(0) == '_' ) {
        //&& SageInterface::get_name(func).at(1) == '_'){
      continue;
    }

    if(SageInterface::get_name(func) == "sc_main")
      continue;

    global_functions_.push_back(new Function(func));
  }
}

void risc::Design::print_design()
{
  std::cout << "Top Modules" << std::endl;
  for(risc::InstanceVectorIter
      iter  = top_modules_.begin();
      iter != top_modules_.end();
      iter++) {

    std::cout << " " << (*iter)->get_ast_type_name()
              << " "
              << (*iter)->get_name()
              << std::endl;

    HierarchicalChannelInstance *channel
      = dynamic_cast<HierarchicalChannelInstance*>(*iter);
    if(channel) {

      channel->get_definition()->print_hierarchy(2);
    } else {

      ModuleInstance *m = dynamic_cast<ModuleInstance*>(*iter);
      m->get_definition()->print_hierarchy(2);
    }
  }
  std::cout << std::endl;

  std::cout << "Modules" << std::endl;
  for(risc::ModuleVectorIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
    (*iter)->print_module_definition();
  }
  std::cout << std::endl;

  std::cout << "Hierarchical channels" << std::endl;
  for(risc::HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
    (*iter)->print_hierarchical_channel_definition();
  }
  std::cout << std::endl;

  std::cout << "Primitive channels" << std::endl;
  for(risc::PrimitiveChannelVectorIter
      iter  = primitive_channel_.begin();
      iter != primitive_channel_.end();
      iter++) {
    std::cout << " " << (*iter)->get_name() << std::endl;
  }
}

void risc::Design::print_all_class_definitions()
{
  for(ModuleVectorConstIter iter = module_definitions_.begin();
      iter != module_definitions_.end(); iter++) {

    std::cout << (*iter)->get_ast_type_name() << std::endl;
  }
}

void risc::Design::print_all_global_functions()
{
  std::cout << std::endl << "Global functions" << std::endl;
  for(risc::FunctionVectorConstIter
      iter  = global_functions_.begin();
      iter != global_functions_.end(); iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
  }
}

risc::Design* risc::GlobalDesign::design_static = NULL;

risc::Design* risc::get_design()
{
  return risc::GlobalDesign::design_static;
}

SgProject* risc::get_project()
{
  return risc::GlobalDesign::design_static->project_;
}

risc::Variable*
risc::Design::find_global_variable(SgVariableDefinition* var_def)
{
  for(VariableVectorIter
      iter  = global_variables_.begin();
      iter != global_variables_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::InstanceTree
risc::Design::get_root()
{
  InstanceTree result;

  if(top_modules_.size() != 1) {
    std::cout << "Number of modules: " << top_modules_.size() << std::endl;
    std::cout << "Only exactly one top module is allowed" << std::endl;
    exit(1);
  }

  InstanceTree::Element top(NULL, top_modules_.at(0));
  result.path_.push_back(top);

  return result;
}

void
risc::Design::reset_instance_counter()
{
  for(ModuleVectorIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {
    (*iter)->instance_counter_ = 0;
  }

  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {
    (*iter)->instance_counter_ = 0;
  }
}

bool
risc::Design::perform_static_analysis()
{
  return dynamic_analysis_filename_.empty();
}

void risc::Design::add_module_to_definition_list(SgClassDefinition *class_def)
{
  Module *m = new Module(class_def);
  module_definitions_.push_back(m);
}

void
risc::Design::add_hierarchical_channel_to_defintion_list(
    SgClassDefinition *class_def)
{
  HierarchicalChannel *hc = new HierarchicalChannel(class_def);
  hierarchical_channel_.push_back(hc);
}

std::string risc::Design::get_design_file_name()
{
  SgFile *file = SageInterface::getEnclosingFileNode(
                     top_modules_.front()->get_ast_node());

  std::string temp = file->get_file_info()->get_filenameString();
  std::size_t found = temp.find_last_of("/\\");

  temp = temp.substr(found+1);

  int size = temp.length();

  if(size >= 2) {
    if(temp.at(size - 2) == '.' &&
        temp.at(size - 1) == 'c') {
      temp = temp.substr(0, size - 2);
    }
  }

  if(size >= 3) {
    if(temp.at(size - 3) == '.' &&
        temp.at(size - 2) == 'c' &&
        temp.at(size - 1) == 'c') {
      temp = temp.substr(0, size - 3);
    }
  }

  if(size >= 4) {
    if(temp.at(size - 4) == '.' &&
        temp.at(size - 3) == 'c' &&
        temp.at(size - 2) == 'p' &&
        temp.at(size - 1) == 'p') {
      temp = temp.substr(0, size - 4);
    }
  }

  return temp;
}

bool risc::Design::is_analyzable()
{
  for(ModuleVectorIter
      module_iter  = module_definitions_.begin();
      module_iter != module_definitions_.end();
      module_iter++) {

    if((*module_iter)->instances_.empty()) {
      continue;
    }

    for(InPortVectorIter
        port_iter  = (*module_iter)->inports_.begin();
        port_iter != (*module_iter)->inports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(InOutPortVectorIter
        port_iter  = (*module_iter)->inoutports_.begin();
        port_iter != (*module_iter)->inoutports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(OutPortVectorIter
        port_iter  = (*module_iter)->outports_.begin();
        port_iter != (*module_iter)->outports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }
  }

  for(HierarchicalChannelVectorIter
      channel_iter  = hierarchical_channel_.begin();
      channel_iter != hierarchical_channel_.end();
      channel_iter++) {

    for(InPortVectorIter
        port_iter  = (*channel_iter)->inports_.begin();
        port_iter != (*channel_iter)->inports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(InOutPortVectorIter
        port_iter  = (*channel_iter)->inoutports_.begin();
        port_iter != (*channel_iter)->inoutports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(OutPortVectorIter
        port_iter  = (*channel_iter)->outports_.begin();
        port_iter != (*channel_iter)->outports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }
  }

  return true;
}

void risc::Design::initialize_interfaces()
{
  SgClassDeclaration *sc_interface
    = tools::SystemCDefinitions::get_sc_interface()->get_declaration();
  sc_interface
    = isSgClassDeclaration(sc_interface->get_firstNondefiningDeclaration());


  const std::vector<SgNode*> &class_defs = tools::RoseNodes::get_class_defs();

  // getting all interfaces
  for(std::vector<SgNode*>::const_iterator
      iter  = class_defs.begin();
      iter != class_defs.end();
      iter++) {

    SgClassDefinition *class_def = isSgClassDefinition(*iter);

    if(class_def->get_inheritances().size() != 1) {
      continue;
    }

    SgClassDeclaration *base_class_decl
      = class_def->get_inheritances().at(0)->get_base_class();
    base_class_decl
      = isSgClassDeclaration(base_class_decl->get_firstNondefiningDeclaration());

    if(base_class_decl == sc_interface) {

      interfaces_.push_back(new Interface(class_def));
    }
  }
}

void risc::Design::initialize_port_mapping()
{
  // Here we setup the port binding for each module.
  for(ModuleVectorIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {
    (*iter)->initialize_mapping();
  }
  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {
    (*iter)->initialize_mapping();
  }
}

void risc::Design::initialize_path_instance_mapper()
{
  if(perform_static_analysis()) {

    // TODO: We activate this when the design takes ParseStatus as argument
    //std::cout << "static analysis" << std::endl;
    path_instance_mapper_ = new StaticPathInstanceMapper(get_root());
  } else {

    // TODO: We activate this when the design takes ParseStatus as argument
    //std::cout << "dynamic analysis" << std::endl;
    path_instance_mapper_
      = new DynamicPathInstanceMapper(dynamic_analysis_filename_);
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
