#include "prediction_event_notification_table.h"

#include<iostream>
#include<cstring>

#include "graph_helper.h"

#include "../internal_representation/cthread.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/hierarchical_channel_instance.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/method.h"
#include "../internal_representation/module.h"
#include "../internal_representation/module_instance.h"
#include "../internal_representation/thread.h"
#include "../internal_representation/path_instance_mapper.h"
#include "../internal_representation/static_path_instance_mapper.h"
#include "../internal_representation/dynamic_path_instance_mapper.h"

risc::sg::PredictionEventNotificationTable::PredictionEventNotificationTable(
    risc::sg::EventConflictTable &ect,
    PathInstanceMapper *path_instance_mapper):
  ect_(ect),
  path_instance_mapper_(path_instance_mapper)
{
  number_of_segments_ = boost::num_vertices(ect_.segment_graph_.graph_);
  max_instances_ = path_instance_mapper_->max_instances_;

  instance_id_to_index_lookup_ = new int*[max_instances_ + 1];

  // We are adding one additional row to safe the segment id
  for(int i=0; i < max_instances_ + 1; i++) {

    instance_id_to_index_lookup_[i] = new int[number_of_segments_];
  }

  // Here, the first row of the instance_id_to_index_lookup table is filled
  // with the segment ids. We have to do that because the segment ids do not
  // incease by one.
  std::pair<VertexIterator, VertexIterator> vp;
  int i = 0;
  for(vp = boost::vertices(ect_.segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    instance_id_to_index_lookup_[0][i]
      = ect_.segment_graph_.graph_[*(vp.first)].id_;
  }

  // We are filling all array fields with -1 to indicate that a specific field
  // is not used. We need that information for the regression tests.
  for(int i = 1; i < max_instances_ + 1; i++) {
    for(int j=0; j < number_of_segments_; j++) {
      instance_id_to_index_lookup_[i][j] = -1;
    }
  }

  // Here, we are filling the instance_id_to_index_lookup_ table. Each field
  // in this table needs an unique value. The value is the associated row/column
  // in the conflict table.
  int index = 0;

  // We are iterating over all the instances in the design
  for(std::vector<PathInstanceMapper::PathWithID>::iterator
      iter  = path_instance_mapper_->path_to_id_mapping_.begin();
      iter != path_instance_mapper_->path_to_id_mapping_.end();
      iter++) {

    // Adding the value of 1 because the first row is reserved for the
    // segment id
    int row = (*iter).second + 1;

    Object *object = (*iter).first.get_object();

    ModuleInstance *module_instance = dynamic_cast<ModuleInstance*>(object);

    Module *module = NULL;

    if(module_instance) {
      module = module_instance->get_definition();
    }

    if(!module) {

      module = dynamic_cast<Module*>((*iter).first.path_.back().first);
    }

    if(module) {

      for(ThreadVectorIter
          thread_iter  = module->threads_.begin();
          thread_iter != module->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            ect_.segment_graph_, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorIter
          cthread_iter  = module->cthreads_.begin();
          cthread_iter != module->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            ect_.segment_graph_, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorIter
          method_iter  = module->methods_.begin();
          method_iter != module->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            ect_.segment_graph_, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }

    // This is the same like for the modules. So, please see the documents above
    HierarchicalChannelInstance *hierarchical_channel_instance
      = dynamic_cast<HierarchicalChannelInstance*>(object);

    if(hierarchical_channel_instance) {

      HierarchicalChannel *hierarchical_channel
        = hierarchical_channel_instance->get_definition();

      for(ThreadVectorIter
          thread_iter  = hierarchical_channel->threads_.begin();
          thread_iter != hierarchical_channel->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            ect_.segment_graph_, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorIter
          cthread_iter  = hierarchical_channel->cthreads_.begin();
          cthread_iter != hierarchical_channel->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            ect_.segment_graph_, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorIter
          method_iter  = hierarchical_channel->methods_.begin();
          method_iter != hierarchical_channel->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            ect_.segment_graph_, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }
  }

  size_of_conflict_table_ = index;


  // Allocating the memory for the VertexDescriptor array
  vertex_lookup_ = new VertexDescriptor[number_of_segments_];

  // Here, we are filling the vertex lookup. The array field at position is the
  // segment with the index instance_id_to_index_lookup_[0][i]
  for(vp = boost::vertices(ect_.segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    int i = 0;
    for(; i < number_of_segments_; i++) {
      if(instance_id_to_index_lookup_[0][i]
          == ect_.segment_graph_.graph_[*(vp.first)].id_) {

        vertex_lookup_[i] = *(vp.first);
        break;
      }
    }
  }

  // creating the conflict table
  combined_table_ = new Time*[size_of_conflict_table_];
  for(int i=0; i<size_of_conflict_table_; i++) {
    combined_table_[i] = new Time[size_of_conflict_table_];
  }

  for(int i=0;i<size_of_conflict_table_;i++) {

    for(int j=0;j<size_of_conflict_table_;j++) {
      combined_table_[i][j].infinite_ = true;
    }
  }

  compute_combined_table();
}

void risc::sg::PredictionEventNotificationTable::print_notification_table()
{
  for(int i = 1; i < max_instances_ + 1; i++) {

    if(i == 1) {

      std::stringstream ss;
      ss << "(Seg, ID)";
      std::string temp = ss.str();
      std::cout << std::setw(9) << temp;
    }

    for(int j=0; j < number_of_segments_; j++) {


      if(instance_id_to_index_lookup_[i][j] == -1) {
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      std::stringstream ss;
      ss << "(" << ids.first << "," << ids.second << ")";
      std::string temp = ss.str();
      std::cout << std::setw(9) << temp;
    }
  }
  std::cout << std::endl;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {


      if(instance_id_to_index_lookup_[i][j] == -1) {
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      std::stringstream ss;
      ss << "(" << ids.first << "," << ids.second << ")";
      std::string temp = ss.str();
      std::cout << std::setw(9) << temp;

      for(int k = 1; k < max_instances_ + 1; k++) {

        for(int l=0; l < number_of_segments_; l++) {

          if(instance_id_to_index_lookup_[k][l] == -1) {
            continue;
          }

          int row = instance_id_to_index_lookup_[i][j];
          int column = instance_id_to_index_lookup_[k][l];

          std::stringstream ss;

          if(combined_table_[row][column].infinite_) {
            ss << "oo:oo";
          } else {
            ss << "(" << combined_table_[row][column].time_units_ << ":"
               << combined_table_[row][column].delta_count_ << ")";
          }

          std::string temp = ss.str();
          std::cout << std::setw(9) << temp;
        }
      }

      std::cout << std::endl;
    }
  }
}

void
risc::sg::PredictionEventNotificationTable::bfs(
    VertexDescriptor start_seg,
    int start_instance)
{
  std::list<Node> current;
  current.push_back(Node(start_seg, Time(0, 0), start_instance));

  int target_index
    = segment_and_instance_id_to_index(
          ect_.segment_graph_.graph_[start_seg].id_, start_instance);

  bool second_iteration = false;

  while(!current.empty()) {

    std::list<Node> next;

    while(!current.empty()) {

      Node node = current.front();
      current.pop_front();

      // following the sg
      if(second_iteration) {

        Graph::in_edge_iterator in_edge_iter_begin, in_edge_iter_end;
        boost::tie(in_edge_iter_begin, in_edge_iter_end)
          = boost::in_edges(node.vd, ect_.segment_graph_.graph_);

        for(; in_edge_iter_begin != in_edge_iter_end; in_edge_iter_begin++) {

          VertexDescriptor source
            = boost::source(*in_edge_iter_begin, ect_.segment_graph_.graph_);

          int source_index
            = segment_and_instance_id_to_index(
                ect_.segment_graph_.graph_[source].id_, node.inst);

          Time waiting_time
            = ect_.segment_graph_.graph_[node.vd].get_min_waiting_time();

          if(combined_table_[source_index][target_index]
              > waiting_time + node.t) {

            combined_table_[source_index][target_index]
              = waiting_time + node.t;

            next.push_back(Node(source,
                  combined_table_[source_index][target_index], node.inst));
          }
        }
      }

      // following the event notification
      int target_index2
        = segment_and_instance_id_to_index(
            ect_.segment_graph_.graph_[node.vd].id_, node.inst);

      for(int source_index = 0;
          source_index < ect_.size_of_conflict_table_;
          source_index++) {

        if(ect_.conflict_table_[source_index][target_index2]) {

          std::pair<int, int> source
            = index_to_segment_and_instance_id(source_index);

          // TODO: This must be optimized!
          std::pair<VertexIterator, VertexIterator> vp;
          for(vp = boost::vertices(ect_.segment_graph_.graph_);
              vp.first != vp.second; ++vp.first) {

            if(source.first == ect_.segment_graph_.graph_[*vp.first].id_) {

              Time temp = node.t;
              if(node.t.time_units_ == 0 ) {
                temp = ect_.segment_graph_.graph_[node.vd].
                          get_min_waiting_time() + node.t;
              }

              if(combined_table_[source_index][target_index] > temp) {

                combined_table_[source_index][target_index] = temp;
                next.push_back(Node(*vp.first, temp,  source.second));
              }

              break;
            }
          }
        }
      }
    }

    second_iteration = true;
    current = next;
  }
}


void
risc::sg::PredictionEventNotificationTable::compute_combined_table()
{
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(ect_.segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    int max_instances
      = get_max_instances(ect_.segment_graph_.graph_[*vp.first].id_);

    for(int instance = 0; instance < max_instances; instance++) {

      int target_index
        = ect_.segment_and_instance_id_to_index(
            ect_.segment_graph_.graph_[*vp.first].id_, instance);

      for(int i = 0; i < ect_.size_of_conflict_table_; i++) {

        if(ect_.conflict_table_[i][target_index]) {
          bfs(*vp.first, instance);
          break;
        }
      }
    }
  }
}


void
risc::sg::PredictionEventNotificationTable::print_notification_table(
    std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i = 1; i < max_instances_ + 1; i++) {

    if(i == 1)
      myfile << "    <th>(Seg,Inst)</th>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {


      if(instance_id_to_index_lookup_[i][j] == -1) {
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <th>"
             << "(" << ids.first << "," << ids.second << ")" << "</th>"
             << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "  <tr>" << std::endl;

      if(instance_id_to_index_lookup_[i][j] == -1) {
        myfile << "  </tr>" << std::endl;
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <td>"
             << "(" << ids.first << "," << ids.second << ")" << "</td>"
             << std::endl;

      for(int k = 1; k < max_instances_ + 1; k++) {

        for(int l=0; l < number_of_segments_; l++) {

          if(instance_id_to_index_lookup_[k][l] == -1) {
            continue;
          }

          int row = instance_id_to_index_lookup_[i][j];
          int column = instance_id_to_index_lookup_[k][l];

          myfile << "    <td> " << combined_table_[row][column]
                 << " </td>" << std::endl;
        }
      }

      myfile << "  </tr>" << std::endl;
    }
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

void
risc::sg::PredictionEventNotificationTable::
print_instance_id_to_index_lookup_table(
    std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i=0; i < number_of_segments_ + 1; i++) {

    if(i == 0) {
      myfile << "    <th> </th>" << std::endl;

    } else {
      myfile << "    <th>"
             << instance_id_to_index_lookup_[0][i-1]
             << "</th>" << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i=1; i < max_instances_ + 1; i++) {

    myfile << "  <tr>" << std::endl;
    myfile << "<td>" << i-1 << "</td>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "<td>" << instance_id_to_index_lookup_[i][j]
             << "</td>" << std::endl;
    }
    myfile << "  </tr>" << std::endl;
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

int
risc::sg::PredictionEventNotificationTable::segment_and_instance_id_to_index(
    int segment_id, int instance_id)
{
  int column_index = -1;
  for(int i = 0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id) {

      column_index = i;
    }
  }

  if(column_index == -1) {
    assert("segment_and_instance_id_to_index: This segment id does not exist"
        && 0);
  }

  return instance_id_to_index_lookup_ [instance_id + 1][column_index];
}

std::pair<int, int>
risc::sg::PredictionEventNotificationTable::index_to_segment_and_instance_id(
    int index)
{
  std::pair<int, int> result;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      if(instance_id_to_index_lookup_[i][j] == index) {

        result.first = instance_id_to_index_lookup_[0][j];
        result.second = i - 1;
        return result;
      }
    }
  }

  std::cout << "index: " << index << std::endl;
  assert("index_to_segment_and_instance_id: Index does not exist" && 0);
}

int risc::sg::PredictionEventNotificationTable::get_max_instances(
    int segment_id)
{
  for(int i=0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id)  {

      int counter = 1;
      for(; counter < max_instances_ + 1; counter++) {
        if(instance_id_to_index_lookup_[counter][i] == -1) {
          break;
        }
      }

      return counter -1;
    }
  }

  assert("Cannot determine max instances" && 0);
  return -1;
}

void
risc::sg::PredictionEventNotificationTable::
print_instance_id_to_index_lookup_table()
{

  for(int i=0; i < max_instances_ + 1; i++) {

    if(i == 0) {

      std::cout << "    ";
    } else {
      std::cout << std::setw(4) << i-1;
    }

    for(int j=0; j < boost::num_vertices(ect_.segment_graph_.graph_); j++) {

      std::cout << std::setw(4) << instance_id_to_index_lookup_[i][j] << " ";
    }

    std::cout << std::endl;
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
