#include "segment.h"

#include "function_annotation_attributes.h"
#include "cached_function_ast_attributes.h"
#include "../internal_representation/design.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/function_helper.h"
#include "../tools/misc.h"
#include "../tools/variable_helper.h"

int risc::sg::Segment::id_counter = 0;

risc::sg::Segment::Segment(
    SgNode const * const boundary_node,
    bool inlined_segment):
  id_(id_counter),
  boundary_node_(boundary_node),
  calls_conflict_function_(false),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched),
  inlined_segment_(inlined_segment),
  deref_pointer_read_(false),
  deref_pointer_write_(false),
  is_channel_segment_(false),
  wait_offset_(-1)
{
  Segment::id_counter++;
  debugging_ = true;

  waits_for_time_ = waits_for_time();
}

risc::sg::Segment::Segment(const Segment &other):
  id_(other.id_),
  boundary_node_(other.boundary_node_),
  expressions_(other.expressions_),
  calls_conflict_function_(other.calls_conflict_function_),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched),
  deref_pointer_read_(other.deref_pointer_read_),
  deref_pointer_write_(other.deref_pointer_write_),
  inlined_segment_(inlined_segment_),
  waits_for_time_(other.waits_for_time_),
  is_channel_segment_(other.is_channel_segment_),
  wait_offset_(other.wait_offset_)
{
  debugging_ = true;
}

bool
risc::sg::Segment::operator==(const Segment &other)
{
  return id_ == other.id_;
}

void
risc::sg::Segment::add_expression(SgNode* node)
{
  if(node) {
#if 0
    // we filter out channel variables
    SgFunctionDefinition *func_def
      = SageInterface::getEnclosingFunctionDefinition(node);
    SgMemberFunctionDeclaration *func_decl
      = isSgMemberFunctionDeclaration(func_def->get_declaration());

    if(func_decl) {
      SgClassDefinition *class_def = func_decl->get_class_scope();
      if(!tools::is_derived_from_sc_prim_channel(class_def)
          && !tools::is_derived_from_sc_channel(class_def)) {
        expressions_.push_back(node);
      } else {

        // we have to maintain event notify function calls for the
        // generation of the event conflict table
        SgFunctionCallExp *func_call = isSgFunctionCallExp(node);
        if(func_call) {

          SgDotExp *dot_exp = isSgDotExp(func_call->get_function());
          if(dot_exp) {

            SgArrowExp *arrow = isSgArrowExp(dot_exp->get_lhs_operand());
            if(arrow) {

              SgVarRefExp *var_ref = isSgVarRefExp(arrow->get_rhs_operand());
              if(var_ref) {

                if(tools::is_sc_event(var_ref)) {

                  SgFunctionDeclaration *func_decl
                    = func_call->getAssociatedFunctionDeclaration();

                  if(SageInterface::get_name(func_decl) == "notify") {
                    expressions_.push_back(node);
                  }
                }
              }
            }
          }
        }
      }
    } else {
      expressions_.push_back(node);
    }
#else
    // add all expressions in a channel to the conflict analysis
    expressions_.push_back(node);
#endif
  }
}

void
risc::sg::Segment::merge(const Segment &other)
{
  if(*this == other)
    return;

  calls_conflict_function_ |= this->calls_conflict_function_;
  deref_pointer_read_ |= this->deref_pointer_read_;
  deref_pointer_write_ |= this->deref_pointer_write_;

  for(Expressions::const_iterator
      iter  = other.expressions_.begin();
      iter != other.expressions_.end();
      iter++) {
    add_expression(*iter);
  }
}

void
risc::sg::Segment::analyze_segment()
{
  for(Segment::Expressions::iterator
      iter  = expressions_.begin();
      iter != expressions_.end();
      iter++) {

    if(isSgExpression(*iter)) {
      SgExpression *expr = isSgExpression(*iter);

      if(isSgAssignOp((*iter)->get_parent())) {
        SgAssignOp *assign_op = isSgAssignOp((*iter)->get_parent());
        if(assign_op->get_lhs_operand() == *iter) {

          analyze_expression(expr, NonRead, NonWrite);
          continue;
        }
      }

      if(isSgReturnStmt(SageInterface::getEnclosingStatement(expr))) {
        analyze_expression(expr, Read, NonWrite);
      } else {
        analyze_expression(expr, NonRead, NonWrite);
      }
    }

    if(isSgExprStatement(*iter)) {
      SgExprStatement *stmt_expr = isSgExprStatement(*iter);
      analyze_expression(stmt_expr->get_expression(), NonRead, NonWrite);
    }

    if(isSgVariableDeclaration(*iter)) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      SgInitializedName *initialized_name = var_decl->get_decl_item(0);

      if(initialized_name->get_initptr()) {

        write_variables_.push_back(
                             isSgVariableSymbol(
                                 initialized_name->
                                     search_for_symbol_from_symbol_table()));

        analyze_expression(initialized_name->get_initptr(), NonRead, NonWrite);
      }
    }
  }
}

void risc::sg::Segment::analyze_expression(
    SgNode *expression, ERead read, EWrite write)
{
  if(expression == NULL) {
    return;
  }

  #define ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(Operator)         \
  if(is##Operator(expression)) {                               \
    Operator *op = is##Operator(expression);                   \
    analyze_expression(op->get_lhs_operand(), Read, NonWrite); \
    analyze_expression(op->get_lhs_operand(), NonRead, Write); \
    analyze_expression(op->get_rhs_operand(), Read, NonWrite); \
    return;                                                    \
  }
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgAndAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgIorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgXorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgLshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgRshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgPlusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMinusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMultAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgDivAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgModAssignOp);
  #undef ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP

  #define ANALYZE_VARS_IN_BINARY_OP(Operator)                  \
  if(is##Operator(expression)) {                               \
    Operator *op = is##Operator(expression);                   \
    analyze_expression(op->get_lhs_operand(), Read, NonWrite); \
    analyze_expression(op->get_rhs_operand(), Read, NonWrite); \
    return;                                                    \
  }
  ANALYZE_VARS_IN_BINARY_OP(SgLessThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLessOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgEqualityOp);
  ANALYZE_VARS_IN_BINARY_OP(SgNotEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLshiftOp);
  ANALYZE_VARS_IN_BINARY_OP(SgRshiftOp);

  ANALYZE_VARS_IN_BINARY_OP(SgBitAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitXorOp);

  ANALYZE_VARS_IN_BINARY_OP(SgAddOp);
  ANALYZE_VARS_IN_BINARY_OP(SgSubtractOp);
  ANALYZE_VARS_IN_BINARY_OP(SgDivideOp);
  ANALYZE_VARS_IN_BINARY_OP(SgMultiplyOp);
  ANALYZE_VARS_IN_BINARY_OP(SgModOp);

  ANALYZE_VARS_IN_BINARY_OP(SgCommaOpExp);
  #undef ANALYZE_VARS_IN_BINARY_OP

  #define ANALYZE_VARS_IN_UNARY_OP(Operator)               \
  if(is##Operator(expression)) {                           \
    Operator *op = is##Operator(expression);               \
    analyze_expression(op->get_operand(), Read, NonWrite); \
    return;                                                \
  }
  ANALYZE_VARS_IN_UNARY_OP(SgUnaryAddOp);
  ANALYZE_VARS_IN_UNARY_OP(SgMinusOp);
  ANALYZE_VARS_IN_UNARY_OP(SgAssignInitializer);
  ANALYZE_VARS_IN_UNARY_OP(SgNotOp);
  ANALYZE_VARS_IN_UNARY_OP(SgBitComplementOp);
  ANALYZE_VARS_IN_UNARY_OP(SgCastExp);
  #undef ANALYZE_VARS_IN_UNARY_OP

  #define ANALYZE_VARS_IN_UNARY_INC_DEC_OP(Operator)    \
  if(is##Operator(expression)) {                        \
    Operator *op = is##Operator(expression);            \
    analyze_expression(op->get_operand(), Read, Write); \
    return;                                             \
  }
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgPlusPlusOp);
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgMinusMinusOp);
  #undef ANALYZE_VARS_IN_UNARY_INC_DEC_OP

  #define ANALYZE_VARS_IN_ACCESS_OP(Operator)               \
  if(is##Operator(expression)) {                            \
    Operator *op = is##Operator(expression);                \
    analyze_expression(op->get_rhs_operand(), Read, write); \
    analyze_expression(op->get_lhs_operand(), Read, write); \
    return;                                                 \
  }
  ANALYZE_VARS_IN_ACCESS_OP(SgDotStarOp);
  ANALYZE_VARS_IN_ACCESS_OP(SgArrowStarOp);
  #undef ANALYZE_VARS_IN_ACCESS_OP



  if(isSgArrowExp(expression)) {
    SgArrowExp *op = isSgArrowExp(expression);

    analyze_expression(op->get_lhs_operand(), Read, write);

    if(write) {
      analyze_expression(op->get_rhs_operand(), read, Write);
    } else {
      analyze_expression(op->get_rhs_operand(), read, NonWrite);
    }
    return;
  }

  if(isSgDotExp(expression)) {
    SgDotExp *op = isSgDotExp(expression);

    if(read && write) {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite);
      analyze_expression(op->get_rhs_operand(), Read, Write);
    } else if(write) {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite);
      analyze_expression(op->get_rhs_operand(), NonRead, Write);
    } else if(read) {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite);
      analyze_expression(op->get_rhs_operand(), Read, NonWrite);
    } else {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite);
      analyze_expression(op->get_rhs_operand(), NonRead, NonWrite);
    }

    return;
  }

  if(isSgSizeOfOp(expression)) {
    SgSizeOfOp *size_of_op = isSgSizeOfOp(expression);
    analyze_expression(size_of_op->get_operand_expr(), Read, NonWrite);
    return;
  }

  if(isSgVarRefExp(expression)) {

    SgVarRefExp *var_ref = isSgVarRefExp(expression);
    risc::tools::analyze_conflict_free(var_ref->get_symbol());

    if(read) {

      read_variables_.push_back(var_ref->get_symbol());
    } else {

      SgModifierType *modifier = isSgModifierType(var_ref->get_type());

      if(modifier) {

        if(modifier->get_typeModifier().get_constVolatileModifier().
            isVolatile()) {

          if(!write) {
            read_variables_.push_back(var_ref->get_symbol());
          }
        }
      }
    }

    if(write) {
      write_variables_.push_back(var_ref->get_symbol());
    }
    return;
  }

  if(isSgFunctionCallExp(expression)) {

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expression);
    SgExpressionPtrList expr_list
      = func_call_exp->get_args()->get_expressions();

    // Here we have to check if any parameter type is a pointer.
    // If so, the segment is with all other segments in conflict
    SgFunctionDeclaration *func_decl
      = tools::get_first_non_defining_function_declaration(func_call_exp);

    if(func_decl) {
      FunctionAnnotationAttributes *function_annotation_attribute =
          dynamic_cast<FunctionAnnotationAttributes*>(
              func_decl->getAttribute("function_annotation"));

      if(function_annotation_attribute) {

        if(!function_annotation_attribute->is_conflict_free_) {
          calls_conflict_function_ = true;

          Conflict conflict(func_call_exp);
          conflicts_.insert(conflict);
        }
      }

      int argument_counter = 0;
      SgExprListExp *arguments = func_call_exp->get_args();
      for(SgExpressionPtrList::iterator
          expr_iter  = expr_list.begin();
          expr_iter != expr_list.end();
          expr_iter++, argument_counter++) {

        if(argument_counter
            < func_decl->get_parameterList()->get_args().size()) {

          SgReferenceType *ref_type
            = isSgReferenceType(func_decl->get_parameterList()
                ->get_args().at(argument_counter)->get_type());

          if(ref_type) {

            analyze_expression(*expr_iter, NonRead, NonWrite);

          } else {

            analyze_expression(*expr_iter, Read, NonWrite);

          }
        } else { // this function signature has the ... parameter

            analyze_expression(*expr_iter, Read, NonWrite);
        }
      }
    }
    return;
  }

  if(isSgPntrArrRefExp(expression)) {
    SgPntrArrRefExp *pointer_array_ref = isSgPntrArrRefExp(expression);

    if(read && write) {
      analyze_expression(pointer_array_ref->get_lhs_operand(), Read, Write);
      analyze_expression(pointer_array_ref->get_rhs_operand(), Read, NonWrite);

    } else if(write) {
      analyze_expression(pointer_array_ref->get_lhs_operand(), NonRead, Write);
      analyze_expression(pointer_array_ref->get_rhs_operand(), Read, NonWrite);

    } else if(read) {
      analyze_expression(pointer_array_ref->get_lhs_operand(), Read, NonWrite);
      analyze_expression(pointer_array_ref->get_rhs_operand(), Read, NonWrite);

    } else {
      analyze_expression(
          pointer_array_ref->get_lhs_operand(), NonRead, NonWrite);
      analyze_expression(pointer_array_ref->get_rhs_operand(), Read, NonWrite);
    }

    return;
  }

  if(isSgAssignOp(expression)) {
    SgAssignOp *assign_op = isSgAssignOp(expression);

    analyze_expression(assign_op->get_lhs_operand(), read, Write);

    analyze_expression(assign_op->get_rhs_operand(), Read, NonWrite);

    return;
  }

  if(isSgConditionalExp(expression)) {
    SgConditionalExp *conditional_exp = isSgConditionalExp(expression);

    analyze_expression(conditional_exp->get_conditional_exp(), Read, NonWrite);

    analyze_expression(conditional_exp->get_true_exp(), Read, NonWrite);

    analyze_expression(conditional_exp->get_false_exp(), Read, NonWrite);

    return;
  }

  if(isSgPointerDerefExp(expression)) {
    SgPointerDerefExp *op = isSgPointerDerefExp(expression);

    analyze_expression(op->get_operand(), Read, NonWrite);

    if(read) {
      deref_pointer_read_ = true;
    }

    if(write) {
      deref_pointer_write_ = true;
    }

    if(debugging_) {
      Conflict conflict(expression);
      conflicts_.insert(conflict);
    }

    return;
  }

  if(isSgAddressOfOp(expression)) {
    SgAddressOfOp *op = isSgAddressOfOp(expression);

    analyze_expression(op->get_operand(), NonRead, NonWrite);

    if(debugging_) {
      Conflict conflict(expression);
      conflicts_.insert(conflict);
    }

    return;
  }

  assert(true && "Operator not supported");
  return;
}

risc::sg::Time
risc::sg::Segment::get_min_waiting_time()
{
  if(!boundary_node_) {
    return Time(0,0);
  }

  const SgStatement *stmt = isSgStatement(boundary_node_);
  const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

  if(expr_stmt) {

    const SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call) {

      if(SageInterface::get_name(
            func_call->getAssociatedFunctionDeclaration()) == "wait") {

        SgExpressionPtrList &function_call_exprs
          = func_call->get_args()->get_expressions();

        // In this case we can wait for
        //   - sc_event_and_list, sc_event_or_list
        //   - sc_event
        //   - sc_zero_time
        //   - (time, time unit)
        //
        if(function_call_exprs.size() == 0) {

          std::cout << "Error: No support of immediate event notification."
                    << std::endl
                    << "       See line "
                    << func_call->get_file_info()->get_line()
                    << " in "
                    << func_call->get_file_info()->get_filename()
                    << std::endl;
          exit(1);

        } else if(function_call_exprs.size() == 1) {

          // SC_ZERO_TIME
          if(func_call->get_args()-> get_expressions().at(0)
              ->unparseToString() == "SC_ZERO_TIME") {

            return risc::sg::Time(0,1);

          } else {

            SgVarRefExp* waiting_event
              = isSgVarRefExp(isSgArrowExp(function_call_exprs.at(0))
                  ->get_rhs_operand());

            // SC_EVENT
            if(risc::tools::is_sc_event(waiting_event) ||
                risc::tools::is_sc_event_and_list(waiting_event) ||
                risc::tools::is_sc_event_or_list(waiting_event)) {

              return risc::sg::Time(0,1);
            }
          }
        } else if(function_call_exprs.size() == 2) {
          // something like wait(1,SC_NS)

          SageInterface::const_int_expr_t waiting_time =
            SageInterface::evaluateConstIntegerExpression(
                isSgCastExp(function_call_exprs.at(0))->get_operand());

          if(waiting_time.hasValue_) {

            return Time(waiting_time.value_ *
                Time::convert_si_magnitude(function_call_exprs.at(1)
                  ->unparseToString()),0);

          } else {

            std::cout << "Error: Wait must have a constant expression."
                      << std::endl;
            std::cout << "       See line "
                      << func_call->get_file_info()->get_line()
                      << " in file "
                      << func_call->get_file_info()->get_filename()
                      << std::endl;

            exit(1);
          }

        } else {

          std::cout << "Error: Waiting time cannot determinted"
                    << " for wait statement ."
                    << std::endl;
          std::cout << "        See line "
                    << func_call->get_file_info()->get_line()
                    << " in file "
                    << func_call->get_file_info()->get_filename()
                    << std::endl;

          exit(1);
        }

        if(func_call->get_args()->get_expressions().size() == 0) {

            return Time(0,0);
        }
      }
    }
  }

  return Time(0,0);
}

bool risc::sg::Segment::waits_for_time()
{
  if(!boundary_node_) {
    return false;
  }

  const SgStatement *stmt = isSgStatement(boundary_node_);
  const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

  if(expr_stmt) {

    const SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call) {

      if(SageInterface::get_name(
            func_call->getAssociatedFunctionDeclaration()) == "wait") {

        SgExpressionPtrList &function_call_exprs
          = func_call->get_args()->get_expressions();

        // In this case we can wait for
        //   - sc_event_and_list, sc_event_or_list
        //   - sc_event
        //   - sc_zero_time
        //   - (time, time unit)
        //
        if(function_call_exprs.size() == 0) {

          return true;

        } else if(function_call_exprs.size() == 1) {

          // SC_ZERO_TIME
          if(func_call->get_args()-> get_expressions().at(0)
              ->unparseToString() == "SC_ZERO_TIME") {

            return true;

          } else {

            SgVarRefExp* waiting_event
              = isSgVarRefExp(isSgArrowExp(function_call_exprs.at(0))
                  ->get_rhs_operand());

            // SC_EVENT
            if(risc::tools::is_sc_event(waiting_event) ||
                risc::tools::is_sc_event_and_list(waiting_event) ||
                risc::tools::is_sc_event_or_list(waiting_event)) {

              return false;
            }
          }
        } else if(function_call_exprs.size() == 2) {
          // something like wait(1,SC_NS)

          SageInterface::const_int_expr_t waiting_time =
            SageInterface::evaluateConstIntegerExpression(
                isSgCastExp(function_call_exprs.at(0))->get_operand());

          if(waiting_time.hasValue_) {

            return true;

          } else {

            std::cout << "Error: Wait must have a constant expression."
                      << std::endl;
            std::cout << "       See line "
                      << func_call->get_file_info()->get_line()
                      << " in file "
                      << func_call->get_file_info()->get_filename()
                      << std::endl;

            exit(1);
          }

        } else {

          std::cout << "Error: Waiting time cannot determinted"
                    << " for wait statement ."
                    << std::endl;
          std::cout << "        See line "
                    << func_call->get_file_info()->get_line()
                    << " in file "
                    << func_call->get_file_info()->get_filename()
                    << std::endl;

          exit(1);
        }
      }
    }
  }

  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
