#ifndef SEGMENT_H_INCLUDED_
#define SEGMENT_H_INCLUDED_

#include<vector>
#include<set>
#include<iostream>

#include "rose.h"
#include "boost/graph/properties.hpp" // needed for color

#include "mapped_variable.h" // needed for Conflict

#include "time.h" // needed for Time

namespace risc {

namespace sg {

class Segment {

public:
  enum ERead {NonRead = 0, Read = 1};
  enum EWrite {NonWrite = 0, Write = 1};

  //typedef std::vector<const SgStatement*> Statements;
  typedef std::vector<SgNode*> Expressions;
  boost::default_color_type color;

  enum Status {Untouched = 0, Touched = 1};
  Status color_for_communication_graph_;
  Status helper_for_graph_algorithms_;

  Expressions expressions_;

  bool debugging_;
  bool inlined_segment_;

  bool is_channel_segment_;
  int wait_offset_;

  std::set<Conflict> conflicts_;

  Segment(SgNode const * const boundary_node, bool inlined_segment);

  Segment(const Segment &other);

  /**
   * \fn       analyze_variable_accesses
   * \brief    This function determines which variable are read and written
   *           in this segment
   */
  void analyze_segment();

  /**
   * \fn       analyze_expression
   * \brief    This function analyzes recursively all read and write accesses
   *           in this expression
   */
  void analyze_expression(SgNode *expression, ERead read, EWrite write);

  friend std::ostream& operator<< (std::ostream &out, const Segment &segment)
  {
    return out;
  }

  /**
   * \fn       add_expression
   * \brief    This function add the given node to the vector of expressions
   */
  void add_expression(SgNode* node);

  /**
   * \fn       operator==
   * \brief    If two segments are equal the return value is true
   */
  bool operator==(const Segment &other);

  /*
   * \brief Merges the other segment in this segment.
   *        No statement will be doubled
   */
  void merge(const Segment &other);

  /**
   * \brief This is the unique id of the segment
   */
  int id_;

  /**
   * \brief This is the associated boundary of the segment
   */
  SgNode const * const boundary_node_;

  /**
   * \brief This static counter will be used to generate id's for the
   *        individual segments.
   */
  static int id_counter;

  /**
   * \brief This color is used for some boost algorithms
   */
  boost::default_color_type color_;

  /**
   * \brief Set of all SgVariableSymbol which are read in this segment
   */
  std::list<SgVariableSymbol*> read_variables_;

  /**
   * \brief Set of all SgVariableSymbol which are written in this segment
   */
  std::list<SgVariableSymbol*> write_variables_;

  /**
   * \brief This segment calls a function with is marked as conflict
   */
  bool calls_conflict_function_;

  /**
   * \fn    get_min_waiting_time
   * \brief This function determines the minimal waiting time for this segment
   */
  Time get_min_waiting_time();


  bool deref_pointer_read_;
  bool deref_pointer_write_;
  bool waits_for_time_;

private:
  /**
   * \fn    waits_for_time
   * \brief This functions determines if a function waits for time
   *        The related flag waits_for_time_ will be set during construction
   */
  bool waits_for_time();
};

} // end of namespace sg

} // end of namespace risc

#endif /* SEGMENT_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
