#include "segment_graph.h"

#include<iostream>
#include<utility>
#include<algorithm>
#include<vector>

#include "boost/graph/breadth_first_search.hpp"

#include "cached_function_ast_attributes.h"
#include "edge_property_writer.h"
#include "function_annotation_attributes.h"
#include "graph_helper.h"
#include "graph_property_writer.h"
#include "port_call_attribute.h"
#include "reference_attributes.h"
#include "vertex_property_writer.h"

#include "lex.function_annotation_parser.h"
#include "parser.function_annotation_parser.tab.h"

#include "../tools/ast_attribute_color.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/function_helper.h"
#include "../tools/misc.h"
#include "../tools/variable_helper.h"

#include "../internal_representation/cthread.h"
#include "../internal_representation/definition.h"
#include "../internal_representation/design.h"
#include "../internal_representation/function.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/method.h"
#include "../internal_representation/module.h"
#include "../internal_representation/thread.h"

int function_annotation_parserparse ();
extern FunctionAnnotation *function_annotation;

risc::sg::SegmentGraph::SegmentGraph(
    FunctionCallBoundarySet function_boundaries):
  function_boundaries_(function_boundaries),
  static_analysis_(risc::get_design()->perform_static_analysis()),
  is_succesfully_created_(true)
{ }

risc::sg::SegmentGraph::SegmentGraph(
    CCxxKeywordsBoundarySet keyword_boundaries):
  keyword_boundaries_(keyword_boundaries),
  static_analysis_(risc::get_design()->perform_static_analysis())
{ }

risc::sg::SegmentGraph::SegmentGraph(
    FunctionCallBoundarySet function_boundaries,
    CCxxKeywordsBoundarySet keyword_boundaries):
  function_boundaries_(function_boundaries),
  keyword_boundaries_(keyword_boundaries),
  static_analysis_(risc::get_design()->perform_static_analysis())
{ }


void
risc::sg::SegmentGraph::clean_graph()
{
  // under certain circumstances expressions can appear twice. For instance,
  // during loop processing etc. These are delete here.
  typedef boost::graph_traits<Graph>::vertex_iterator VertexIterator;
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(graph_);
      vp.first != vp.second;
      ++vp.first) {

    Segment::Expressions &expressions = graph_[*vp.first].expressions_;
    std::set<SgNode*> found_nodes;
    for(size_t index = 0; index != expressions.size();) {

      if(found_nodes.find(expressions[index]) == found_nodes.end()) {
        found_nodes.insert(expressions[index]);
        ++index;
      } else {
        expressions.erase(expressions.begin() + index);
      }
    }
  }

  // Deleting all pre cached function information
  for(std::set<VertexDescriptor>::iterator
      precached_vertex_iter  = precached_function_segments_.begin();
      precached_vertex_iter != precached_function_segments_.end();
      precached_vertex_iter++) {

    boost::clear_vertex(*precached_vertex_iter, graph_);
    boost::remove_vertex(*precached_vertex_iter, graph_);
  }
}

void
risc::sg::SegmentGraph::print_graph(std::string filename)
{
  VertexPropertyWriter::print_read_write_access = false;

  std::ofstream file;
  file.open(filename.c_str(), std::ofstream::out | std::ofstream::trunc);

  boost::write_graphviz(
             file,
             graph_,
             VertexPropertyWriter(graph_),
             EdgePropertyWriter(graph_),
             GraphPropertyWriter(graph_),
             boost::get(&Segment::id_, graph_));

  file.close();
}

void
risc::sg::SegmentGraph::print_graph_read_write_access(std::string filename)
{
  VertexPropertyWriter::print_read_write_access = true;

  std::ofstream file;
  file.open(filename.c_str(), std::ofstream::out | std::ofstream::trunc);

  boost::write_graphviz(
             file,
             graph_,
             VertexPropertyWriter(graph_),
             EdgePropertyWriter(graph_),
             GraphPropertyWriter(graph_),
             boost::get(&Segment::id_, graph_));

  file.close();
}

void
risc::sg::SegmentGraph::add_function(Function *function,
    bool duplicate_segments)
{
  int has_instance = false;
  if(function->parent_->type_ == Definition::ModuleType) {
    Module *m = dynamic_cast<Module*>(function->parent_);
    has_instance = m->instance_counter_ != 0;
  }

  if(function->parent_->type_ == Definition::HierarchicalChannelType) {
    Module *m = dynamic_cast<HierarchicalChannel*>(function->parent_);
    has_instance = m->instance_counter_ != 0;
  }

  if(!has_instance) {
    std::cout << "Warning: The segment graph can only be generated for"
              << " threads where the hosting module/channel has at least"
              << " one instance."
              << std::endl;
    std::cout << "         Skipping function "
              << function->get_ast_node()->get_declaration()
                   ->get_qualified_name().getString()
              << std::endl;

    is_succesfully_created_ = false;
    return;
  }

  Segment first_segment(NULL, false);

  VertexDescriptor segment_vertex = boost::add_vertex(first_segment, graph_);

  SegmentSet current_segments;
  SegmentSet break_segments;
  SegmentSet continue_segments;

  function->starting_id_ = graph_[segment_vertex].id_;

  current_segments.insert(segment_vertex);

  build_graph(function->get_ast_node()->get_body(), current_segments,
      break_segments, continue_segments, duplicate_segments);

  handle_recursive_calls();
}

void
risc::sg::SegmentGraph::add_expression_to_segment(
    SegmentSet segments,
    SgNode *expr)
{
  for(SegmentSet::iterator
      segment_iter  = segments.begin();
      segment_iter != segments.end();
      segment_iter++) {

    graph_[*segment_iter].add_expression(expr);
  }
}

risc::sg::SegmentGraph::SegmentSet
risc::sg::SegmentGraph::duplicate_empty_segments(SegmentSet segments)
{
  SegmentSet duplicates;

  for(SegmentSet::iterator
      iter_segment  = segments.begin();
      iter_segment != segments.end();
      iter_segment++){

    Segment duplicate(graph_[*iter_segment].boundary_node_, false);

    VertexDescriptor duplicated_vertex =
      boost::add_vertex(duplicate, graph_);

    Graph::in_edge_iterator in_edge_iter_begin, in_edge_iter_end;

    boost::tie(in_edge_iter_begin, in_edge_iter_end) =
      boost::in_edges((*iter_segment), graph_);

    for(; in_edge_iter_begin != in_edge_iter_end; in_edge_iter_begin++) {

      boost::add_edge(
                 boost::source(*in_edge_iter_begin, graph_),
                 duplicated_vertex, graph_ );
    }

    duplicates.insert(duplicated_vertex);
  }

  return duplicates;
}

risc::sg::SegmentGraph::SegmentSet
risc::sg::SegmentGraph::duplicate_segments(SegmentSet segments)
{
  SegmentSet duplicates;

  for(SegmentSet::iterator
      iter_segment  = segments.begin();
      iter_segment != segments.end();
      iter_segment++){

    Segment duplicate(graph_[*iter_segment]);

    VertexDescriptor duplicated_vertex =
      boost::add_vertex(duplicate, graph_);

    Graph::in_edge_iterator in_edge_iter_begin, in_edge_iter_end;

    boost::tie(in_edge_iter_begin, in_edge_iter_end) =
      boost::in_edges((*iter_segment), graph_);

    for(; in_edge_iter_begin != in_edge_iter_end; in_edge_iter_begin++) {

      boost::add_edge(
                 boost::source(*in_edge_iter_begin, graph_),
                 duplicated_vertex, graph_ );
    }

    duplicates.insert(duplicated_vertex);
  }

  return duplicates;
}

risc::sg::SegmentGraph::SegmentSet
risc::sg::SegmentGraph::build_graph(
     SgStatement *current_stmt,
     SegmentSet current_segments,
     SegmentSet &break_segments,
     SegmentSet &continue_segments,
     bool duplicate_segments)
{
  // null pointer protection
  if(current_stmt == NULL)
    return current_segments;

  // we have to figure out which kind of statement we are addressing
  if(is_boundary_stmt(current_stmt)) { //Special case of a boundary statement

    if(V_SgSwitchStatement == current_stmt->variantT()) {
      SgSwitchStatement *switch_stmt = isSgSwitchStatement(current_stmt);

      // leaving segment
      Segment leaving_segment(NULL, false);
      VertexDescriptor leaving_vertex
        = boost::add_vertex(leaving_segment, graph_);

      SegmentSet leaving_segments;
      leaving_segments.insert(leaving_vertex);

      // we have a separated break segment set that we don't overwrite the
      // break segments for other loops or switch statements
      SegmentSet break_segments;

      // here we are analyzing the selector which will be added to the current
      // segments
      current_segments = build_graph(switch_stmt->get_item_selector(),
                             current_segments,
                             break_segments,
                             continue_segments,
                             duplicate_segments);

      std::vector<SgStatement*> switch_cases
        = SageInterface::getSwitchCases(switch_stmt);

      // This flag will tell us if we need a edge from the segments before
      // the switch to the segments after the switch
      bool default_cases_exists = false;

      // here we analyze each switch case
      for(std::vector<SgStatement*>::iterator
          case_iter  = switch_cases.begin();
          case_iter != switch_cases.end();
          case_iter++) {

        SgCaseOptionStmt *switch_case = isSgCaseOptionStmt(*case_iter);
        SgDefaultOptionStmt *default_case = isSgDefaultOptionStmt(*case_iter);
        SgBasicBlock *bb = NULL;

        if(switch_case)
          bb = isSgBasicBlock(switch_case->get_body());

        if(default_case) {
          bb = isSgBasicBlock(default_case->get_body());
          default_cases_exists = true;
        }

        // the beginning segment for this case
        Segment case_segment(NULL, false);
        VertexDescriptor case_vertex
          = boost::add_vertex(case_segment, graph_);
        SegmentSet case_segments;
        case_segments.insert(case_vertex);

        // we are linking the new segment for this case with the segment
        // current segments
        for(SegmentSet::iterator
            iter  = current_segments.begin();
            iter != current_segments.end();
            iter++) {

          boost::add_edge(*iter, case_vertex, graph_);
        }

        // analyzing this case
        SegmentSet break_segments;
        SegmentSet result = build_graph(bb,
                                case_segments,
                                break_segments,
                                continue_segments,
                                duplicate_segments);

        // if the segment set result is empty, all control flow path will end
        // in a break statement, so we can analyze the next case
        if(!result.empty()) {

          // However, if result is not empty the next case will be executed as
          // well. So, here we are iterating case by case until we hit the first
          // case where all control flow path end in break statements
          for(std::vector<SgStatement*>::iterator
              following_segment_iter  = case_iter + 1;
              following_segment_iter != switch_cases.end();
              following_segment_iter++) {

            switch_case = isSgCaseOptionStmt(*following_segment_iter);
            default_case = isSgDefaultOptionStmt(*following_segment_iter);
            bb = NULL;

            if(switch_case)
              bb = isSgBasicBlock(switch_case->get_body());

            if(default_case)
              bb = isSgBasicBlock(default_case->get_body());

            result = build_graph(bb, result, break_segments, result,
                         duplicate_segments);

            if(result.empty()) {
              break;
            }
          }
        }

        // connecting all break segments to the leaving vertex
        for(SegmentSet::iterator
            iter  = break_segments.begin();
            iter != break_segments.end();
            iter++) {

          boost::add_edge(*iter, leaving_vertex, graph_);
        }

        // connecting all result segments to the leaving vertex
        for(SegmentSet::iterator
            iter  = result.begin();
            iter != result.end();
            iter++) {

          boost::add_edge(*iter, leaving_vertex, graph_);
        }
      }

      // If we don't have any default case, it is possible to skip the entire
      // switch statement
      if(!default_cases_exists) {

        // So, we have to a transitions from all current segments to the leaving
        // segment
        for(SegmentSet::iterator
            iter  = current_segments.begin();
            iter != current_segments.end();
            iter++) {

          boost::add_edge(*iter, leaving_vertex, graph_);
        }
      }

      return leaving_segments;

    } else if(V_SgDoWhileStmt == current_stmt->variantT()) {

      SgDoWhileStmt *do_while_stmt = isSgDoWhileStmt(current_stmt);
      SgExprStatement *expr_stmt
        = isSgExprStatement(do_while_stmt->get_condition());

      // Segment for the loop condition
      Segment loop_condition_segment(NULL, false);
      VertexDescriptor loop_condition_vertex =
        boost::add_vertex(loop_condition_segment, graph_);

      SegmentSet loop_condition_segments;
      loop_condition_segments.insert(loop_condition_vertex);

      add_expression_to_segment(loop_condition_segments,
          expr_stmt->get_expression());

      // Segment for the loop body
      Segment loop_body(NULL, false);
      VertexDescriptor loop_body_vertex = boost::add_vertex(loop_body, graph_);

      // linking the current segments to the first segment of the loop body
      // it a do while loop
      for(SegmentSet::iterator
          iter  = current_segments.begin();
          iter != current_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_body_vertex, graph_);
      }

      SegmentSet loop_body_segments;
      loop_body_segments.insert(loop_body_vertex);

      // linking the loop condition to the first segment of the loop body
      for(SegmentSet::iterator
          iter  = loop_condition_segments.begin();
          iter != loop_condition_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_body_vertex, graph_);
      }

      // We have new segment sets for the break and continue segments that we
      // don't overwrite these from a potential outer loop
      SegmentSet break_segments, continue_segments;

      // analyzing the loop body
      SegmentSet loop_segments = build_graph(do_while_stmt->get_body(),
                                     loop_body_segments, break_segments,
                                     continue_segments,
                                     duplicate_segments);

      // linking back to the condition
      for(SegmentSet::iterator
          iter  = loop_segments.begin();
          iter != loop_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // linking back to the condition
      for(SegmentSet::iterator
          iter  = continue_segments.begin();
          iter != continue_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // if the loop condition vertex has no incoming edges we can delete it
      Graph::in_edge_iterator in_edge_begin, in_edge_end, in_edge_iter;
      boost::tie(in_edge_begin, in_edge_end)
        = boost::in_edges(loop_condition_vertex, graph_);

      if(in_edge_begin == in_edge_end) {

        boost::clear_vertex(loop_condition_vertex, graph_);
        boost::remove_vertex(loop_condition_vertex, graph_);
      }

      // leaving segment
      Segment leaving_body(NULL, false);
      VertexDescriptor leaving_vertex = boost::add_vertex(leaving_body, graph_);

      SegmentSet leaving_segments;
      leaving_segments.insert(leaving_vertex);

      for(SegmentSet::iterator
          iter_leaving  = leaving_segments.begin();
          iter_leaving != leaving_segments.end();
          iter_leaving++) {

        for(SegmentSet::iterator
            iter_loop_condition  = loop_condition_segments.begin();
            iter_loop_condition != loop_condition_segments.end();
            iter_loop_condition++) {

          if(in_edge_begin != in_edge_end) {
            boost::add_edge(*iter_loop_condition, *iter_leaving, graph_);
          }
        }

        for(SegmentSet::iterator
            iter_break  = break_segments.begin();
            iter_break != break_segments.end();
            iter_break++) {
          boost::add_edge(*iter_break, *iter_leaving, graph_);
        }
      }

      return leaving_segments;

    } else if(V_SgForStatement == current_stmt->variantT()) {

      SgForStatement *for_stmt = isSgForStatement(current_stmt);
      SgExprStatement *expr_stmt = isSgExprStatement(for_stmt->get_test());

      // Segment for the loop condition
      Segment loop_condition_segment(NULL, false);
      VertexDescriptor loop_condition_vertex =
        boost::add_vertex(loop_condition_segment, graph_);

      SegmentSet loop_condition_segments;
      loop_condition_segments.insert(loop_condition_vertex);

      add_expression_to_segment(loop_condition_segments,
          expr_stmt->get_expression());

      // here we are adding the linking between the current segments and the
      // condition segment
      for(SegmentSet::iterator
          iter  = current_segments.begin();
          iter != current_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // Adding the init expression/declaration to the current segments
      // We do not have an extra segment for that.
      for (SgStatementPtrList::iterator
          init_iter  = for_stmt->get_init_stmt().begin();
          init_iter != for_stmt->get_init_stmt().end();
          init_iter++) {

        add_expression_to_segment(current_segments, *init_iter);
      }

      // Segment for the loop body
      Segment loop_body(NULL, false);
      VertexDescriptor loop_body_vertex = boost::add_vertex(loop_body, graph_);

      SegmentSet loop_body_segments;
      loop_body_segments.insert(loop_body_vertex);

      // linking of the condition segment and the first segment of the loop
      // body
      for(SegmentSet::iterator
          iter  = loop_condition_segments.begin();
          iter != loop_condition_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_body_vertex, graph_);
      }

      SegmentSet break_segments, continue_segments;

      // analyzing of the loop body
      SegmentSet loop_segments = build_graph(for_stmt->get_loop_body(),
                                     loop_body_segments, break_segments,
                                     continue_segments,
                                     duplicate_segments);

      // adding the increment step
      add_expression_to_segment(continue_segments, for_stmt->get_increment());
      add_expression_to_segment(loop_segments, for_stmt->get_increment());

      // linking back to the condition
      for(SegmentSet::iterator
          iter  = loop_segments.begin();
          iter != loop_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // linking back to the condition
      for(SegmentSet::iterator
          iter  = continue_segments.begin();
          iter != continue_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // leaving segment, with this segment the flow will continue
      Segment leaving_body(NULL, false);
      VertexDescriptor leaving_vertex = boost::add_vertex(leaving_body, graph_);

      SegmentSet leaving_segments;
      leaving_segments.insert(leaving_vertex);

      // we have to link the condition and the break segments to the leaving
      // segment
      for(SegmentSet::iterator
          iter_leaving  = leaving_segments.begin();
          iter_leaving != leaving_segments.end();
          iter_leaving++) {

        for(SegmentSet::iterator
            iter_loop_condition  = loop_condition_segments.begin();
            iter_loop_condition != loop_condition_segments.end();
            iter_loop_condition++) {

          boost::add_edge(*iter_loop_condition, *iter_leaving, graph_);
        }

        for(SegmentSet::iterator
            iter_break  = break_segments.begin();
            iter_break != break_segments.end();
            iter_break++) {
          boost::add_edge(*iter_break, *iter_leaving, graph_);
        }
      }

      return leaving_segments;

    } else if(V_SgWhileStmt == current_stmt->variantT()) {

      SgWhileStmt *while_stmt = isSgWhileStmt(current_stmt);
      SgExprStatement *expr_stmt
        = isSgExprStatement(while_stmt->get_condition());

      // Segment for the loop condition
      Segment loop_condition_segment(NULL, false);
      VertexDescriptor loop_condition_vertex =
        boost::add_vertex(loop_condition_segment, graph_);

      SegmentSet loop_condition_segments;
      loop_condition_segments.insert(loop_condition_vertex);

      // adding the loop condition to the loop condition segment
      add_expression_to_segment(loop_condition_segments,
          expr_stmt->get_expression());

      // We are linking all current segments to the loop condition segment
      for(SegmentSet::iterator
          iter  = current_segments.begin();
          iter != current_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // Segment for the loop body
      Segment loop_body(NULL, false);
      VertexDescriptor loop_body_vertex = boost::add_vertex(loop_body, graph_);

      SegmentSet loop_body_segments;
      loop_body_segments.insert(loop_body_vertex);

      // linking the loop condition segment to the first segment of the
      // loop
      for(SegmentSet::iterator
          iter  = loop_condition_segments.begin();
          iter != loop_condition_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_body_vertex, graph_);
      }

      SegmentSet break_segments, continue_segments;

      // analyzing of the loop body
      SegmentSet loop_segments = build_graph(while_stmt->get_body(),
                                     loop_body_segments,
                                     break_segments,
                                     continue_segments,
                                     duplicate_segments);

      // linking back to the condition
      for(SegmentSet::iterator
          iter  = loop_segments.begin();
          iter != loop_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // linking back to the condition
      for(SegmentSet::iterator
          iter  = continue_segments.begin();
          iter != continue_segments.end();
          iter++) {

        boost::add_edge(*iter, loop_condition_vertex, graph_);
      }

      // leaving segment, with this segment the flow will continue
      Segment leaving_body(NULL, false);
      VertexDescriptor leaving_vertex = boost::add_vertex(leaving_body, graph_);

      SegmentSet leaving_segments;
      leaving_segments.insert(leaving_vertex);

      // we have to link the condition and the break segments to the leaving
      // segment
      for(SegmentSet::iterator
          iter_leaving  = leaving_segments.begin();
          iter_leaving != leaving_segments.end();
          iter_leaving++) {

        for(SegmentSet::iterator
            iter_loop_condition  = loop_condition_segments.begin();
            iter_loop_condition != loop_condition_segments.end();
            iter_loop_condition++) {

          boost::add_edge(*iter_loop_condition, *iter_leaving, graph_);
        }

        for(SegmentSet::iterator
            iter_break  = break_segments.begin();
            iter_break != break_segments.end();
            iter_break++) {
          boost::add_edge(*iter_break, *iter_leaving, graph_);
        }
      }

      return leaving_segments;

    } else if(V_SgIfStmt == current_stmt->variantT()){

      SgIfStmt *if_stmt = isSgIfStmt(current_stmt);
      SgExprStatement *expr_stmt
        = isSgExprStatement(if_stmt->get_conditional());
      add_expression_to_segment(current_segments,
          expr_stmt->get_expression());

      Segment true_body_segment(NULL, false);
      Segment joining_segment(NULL, false);

      VertexDescriptor true_body_vertex
        = boost::add_vertex(true_body_segment, graph_);

      SegmentSet true_body_segments;
      true_body_segments.insert(true_body_vertex);

      for(SegmentSet::iterator
          iter  = current_segments.begin();
          iter != current_segments.end();
          iter++) {

        boost::add_edge(*iter,true_body_vertex,graph_);
      }

      SegmentSet break_segments_1, continue_segments_1;
      SegmentSet temp1
        = build_graph(if_stmt->get_true_body(), true_body_segments,
              break_segments_1, continue_segments_1, duplicate_segments);

      break_segments.insert(break_segments_1.begin(), break_segments_1.end());
      continue_segments.insert(continue_segments_1.begin(),
                            continue_segments_1.end());

      VertexDescriptor joining_vertex
        = boost::add_vertex(joining_segment, graph_);
      SegmentSet joining_segments;

      if(if_stmt->get_false_body()) {

        Segment false_body_segment(NULL, false);
        VertexDescriptor false_body_vertex
          = boost::add_vertex(false_body_segment, graph_);
        SegmentSet false_body_segments;
        false_body_segments.insert(false_body_vertex);
        SegmentSet break_segments_2, continue_segments_2;
        SegmentSet temp2 = build_graph(if_stmt->get_false_body(),
            false_body_segments, break_segments_2, continue_segments_2,
            duplicate_segments);

        break_segments.insert(break_segments_2.begin(), break_segments_2.end());
        continue_segments.insert(continue_segments_2.begin(),
                              continue_segments_2.end());
        for(SegmentSet::iterator
            iter  = current_segments.begin();
            iter != current_segments.end();
            iter++) {

          boost::add_edge(*iter,false_body_vertex,graph_);
        }
        for(SegmentSet::iterator
            iter  = temp2.begin();
            iter != temp2.end();
            iter++) {

          boost::add_edge(*iter,joining_vertex,graph_);
        }
      } else {
        for(SegmentSet::iterator
            iter  = current_segments.begin();
            iter != current_segments.end();
            iter++) {

          boost::add_edge(*iter,joining_vertex,graph_);
        }
      }

      for(SegmentSet::iterator
          iter  = temp1.begin();
          iter != temp1.end();
          iter++) {

        boost::add_edge(*iter,joining_vertex,graph_);
      }

      joining_segments.insert(joining_vertex);

      return joining_segments;
    }

    // Here we add the boundary to all the current segments
    //add_statement_to_segment(current_segments, current_stmt);

    if(!duplicate_segments) {
      // here we check if this segment already exists
      typedef boost::graph_traits<Graph>::vertex_iterator VertexIterator;
      std::pair<VertexIterator, VertexIterator> vp;
      for(vp = boost::vertices(graph_);
          vp.first != vp.second;
          ++vp.first) {

        if(graph_[*vp.first].boundary_node_ == current_stmt) {

          // All the new edges are predecessors of the new segment
          for(SegmentSet::iterator
              segment_iter  = current_segments.begin();
              segment_iter != current_segments.end();
              segment_iter++) {

            if(!boost::edge(*segment_iter, *vp.first, graph_).second) {

              boost::add_edge(*segment_iter, *vp.first, graph_);
            }
          }

          // If this segment already exists, we can return an empty segment set.
          // In the calling basic block loop, the loop will break and stop
          // analyzing the same segment twice.
          SegmentSet return_segments;
          return return_segments;
        }
      }
    }

    // After a boundary a new segment starts.
    // Creating a new segment for the new part after the boundary
    Segment new_segment(current_stmt, false);

    // Adding the new segment to the graph
    VertexDescriptor new_seg_vertex = boost::add_vertex(new_segment, graph_);

    // All the new edges are predecessors of the new segment
    for(SegmentSet::iterator
        segment_iter  = current_segments.begin();
        segment_iter != current_segments.end();
        segment_iter++) {

      if(!boost::edge(*segment_iter, new_seg_vertex, graph_).second) {
        boost::add_edge(*segment_iter, new_seg_vertex, graph_);
      }
    }

    SegmentSet return_segments;
    return_segments.insert(new_seg_vertex);
    return return_segments;

  } else if(isSgIfStmt(current_stmt)) { // if else statement
    SgIfStmt *if_stmt = isSgIfStmt(current_stmt);

    // add the condition expression to the current segments
    SgExprStatement *expr_stmt = isSgExprStatement(if_stmt->get_conditional());
    add_expression_to_segment(current_segments, expr_stmt->get_expression());

    SegmentSet break_segments_1, break_segments_2;
    SegmentSet continue_segments_1, continue_segments_2;

    SegmentSet temp1 = build_graph(if_stmt->get_true_body(),
                           current_segments,
                           break_segments_1,
                           continue_segments_1,
                           duplicate_segments);

    SegmentSet temp2 = build_graph(if_stmt->get_false_body(),
                           current_segments,
                           break_segments_2,
                           continue_segments_2,
                           duplicate_segments);

    break_segments.insert(break_segments_1.begin(), break_segments_1.end());
    break_segments.insert(break_segments_2.begin(), break_segments_2.end());

    continue_segments.insert(continue_segments_1.begin(),
                          continue_segments_1.end());
    continue_segments.insert(continue_segments_2.begin(),
                          continue_segments_2.end());

    temp1.insert(temp2.begin(), temp2.end());

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(expr_stmt->get_expression(),
        current_segments);

    return temp1;

  } else if(isSgWhileStmt(current_stmt)) {

    // This example will be used to explain the following code.
    // The vertex descriptor 'loop_vertex' is a temporary segment graph node
    // for the loop which will be deleted after the loop processing. This node
    // includes all the statements until the boundary statements. In the given
    // example this would be 'int bb'.
    //
    // int aa;  <- current segments
    // while(condition) {
    //   int bb;  <- loop_vertex starts here
    //   wait();
    //   int cc;  <- boost::target(*out_edge_iter, graph_),
    //   wait();
    //   int dd;  <- leaf_segments_of_loop
    // }
    // int ee;
    // wait();
    // int ff;

    SgWhileStmt *while_stmt = isSgWhileStmt(current_stmt);
    SgExpression *while_condition =
      isSgExprStatement(while_stmt->get_condition())->get_expression();

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(while_condition,
        current_segments);

    // add the condition expression to the current segments
    add_expression_to_segment(current_segments, while_condition);

    // Here we create a new segment for the loop body
    Segment loop_segment(NULL, false);
    VertexDescriptor loop_vertex = boost::add_vertex(loop_segment, graph_);
    SegmentSet leaf_segments_of_loop;
    leaf_segments_of_loop.insert(loop_vertex);

    // Analyze the loop body
    SegmentSet new_current_segments, break_segments, continue_segments;
    leaf_segments_of_loop = build_graph(while_stmt->get_body(),
                                leaf_segments_of_loop,
                                break_segments,
                                continue_segments,
                                duplicate_segments);

    // We are adding the edges among the individual segments
    insert_loop_edges(loop_vertex,
        current_segments,
        continue_segments,
        leaf_segments_of_loop);

    // Here we are adding all required expression to the
    // 'leaf_segments_of_the_loop'
    for(SegmentSet::iterator
        leaf_iter  = leaf_segments_of_loop.begin();
        leaf_iter != leaf_segments_of_loop.end();
        leaf_iter++) {

      // Here we add to all 'leaf_segments_of_loop' the condition in the while
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'condition'. Afterwards, it will look like
      // 'int dd; condition;'.
      graph_[*leaf_iter].add_expression(while_condition);

      // Here we add all the expressions of the 'loop_vertex' (int bb)
      // at the end of the last segment. So at this point the segment
      // 'int dd' has changed into 'int dd; condition; int bb;
      graph_[*leaf_iter].merge(graph_[loop_vertex]);
    }

    // Here we are adding all required expression to the
    // 'continue_segments'
    for(SegmentSet::iterator
        continue_segments_iter  = continue_segments.begin();
        continue_segments_iter != continue_segments.end();
        continue_segments_iter++) {

      // Here we add to all 'continue_segments' the condition in the while
      // loop.
      graph_[*continue_segments_iter].add_expression(while_condition);

      // Here we add all the expressions of the 'loop_vertex' (int bb)
      // at the end of the last segment.
      graph_[*continue_segments_iter].merge(graph_[loop_vertex]);

      // This covers the case that we delete the vertex 'loop_vertex' later,
      // however, the loop_vertex is also a in 'leaf_segments_of_loop'. So, we
      // have to consider the merged 'current_segment' as leaving segment as
      // well.
      if(leaf_segments_of_loop.find(loop_vertex) !=
          leaf_segments_of_loop.end()) {

        new_current_segments.insert(*continue_segments_iter);
      }
    }

    for(SegmentSet::iterator
        current_segments_iter  = current_segments.begin();
        current_segments_iter != current_segments.end();
        current_segments_iter++) {

      // Here, we merge the current segments like 'int aa' with the first
      // segment of the loop. At the end, we can delete the first node of the
      // loop because we are merge it here.
      graph_[*current_segments_iter].merge(graph_[loop_vertex]);

      // This covers the case that we delete the vertex 'loop_vertex' later,
      // however, the loop_vertex is also in 'current_segments'. So, we have
      // to consider the merged 'current_segment' as leaving segment as well.
      if(leaf_segments_of_loop.find(loop_vertex) !=
          leaf_segments_of_loop.end()) {

        new_current_segments.insert(*current_segments_iter);
      }
    }

    // In the following we add the 'leaf_segments_of_loop', the
    // 'current_segments', the 'break_segments', and the 'continue_segments'
    // to the 'new_current_segments'. These are the segments which will leave
    // the loop.
    new_current_segments.insert(leaf_segments_of_loop.begin(),
                             leaf_segments_of_loop.end());
    new_current_segments.insert(current_segments.begin(),
                             current_segments.end());
    new_current_segments.insert(break_segments.begin(),
                             break_segments.end());
    new_current_segments.insert(continue_segments.begin(),
                             continue_segments.end());

    // Here we remove the loop vertex from the graph because all the
    // statements already moved into the current segments.
    new_current_segments.erase(loop_vertex);
    boost::clear_vertex(loop_vertex, graph_);
    boost::remove_vertex(loop_vertex, graph_);

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(while_condition,
        new_current_segments);

    return new_current_segments;

  } else if(isSgForStatement(current_stmt)) { // for loop statement

    // This example will be used to explain the following code.
    // The vertex descriptor 'loop_vertex' is a temporary segment graph node
    // for the loop which will be deleted after the loop processing. This node
    // includes all the statements until the boundary statements. In the given
    // example this would be 'int bb'.
    //
    // int aa;  <- current segments
    // for(int i = 0; i < 99; i++) {
    //   int bb;  <- loop_vertex starts here
    //   wait();
    //   int cc;  <- boost::target(*out_edge_iter, graph_),
    //   wait();
    //   int dd;  <- leaf_segments_of_loop
    // }
    // int ee;
    // wait();
    // int ff;

    SgForStatement *for_stmt = isSgForStatement(current_stmt);
    SgExpression *for_condition = NULL;

    if(for_stmt->get_test()) {
      SgExprStatement *expr_stmt = isSgExprStatement(for_stmt->get_test());
      if(expr_stmt) {
        for_condition = expr_stmt->get_expression();
      }
    }

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    for(SgStatementPtrList::const_iterator
        iter  = for_stmt->get_init_stmt().begin();
        iter != for_stmt->get_init_stmt().end();
        iter++ ) {
      mark_not_conflict_free_func_calls_in_segment(*iter, current_segments);
    }


    // add the initialization expression to the current segments
    for(SgStatementPtrList::const_iterator
        iter  = for_stmt->get_init_stmt().begin();
        iter != for_stmt->get_init_stmt().end();
        iter++ ) {

      add_expression_to_segment(current_segments, *iter);

      // Here, we check if there is any function call in the expression which
      // is annotated as not-conflict-free
      //mark_not_conflict_free_func_calls_in_segment(*iter, current_segments);
    }

    // add the condition expression to the current segments
    add_expression_to_segment(current_segments, for_condition);

    // Here we create a new segment for the loop body
    Segment loop_segment(NULL, false);
    VertexDescriptor loop_vertex = boost::add_vertex(loop_segment, graph_);
    SegmentSet leaf_segments_of_loop;
    leaf_segments_of_loop.insert(loop_vertex);

    // Analyze the loop body
    SegmentSet break_segments, continue_segments, new_current_segments;
    leaf_segments_of_loop = build_graph(for_stmt->get_loop_body(),
                                leaf_segments_of_loop,
                                break_segments,
                                continue_segments,
                                duplicate_segments);


    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    // We take the leaf_segments_of_loop because they lead to a new loop
    // iteration
    mark_not_conflict_free_func_calls_in_segment(for_stmt->get_increment(),
        leaf_segments_of_loop);

    // We are adding the edges among the individual segments
    insert_loop_edges(loop_vertex,
        current_segments,
        continue_segments,
        leaf_segments_of_loop);

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(for_stmt->get_increment(),
        continue_segments);

    // Here we are adding all required expression to the
    // 'leaf_segments_of_the_loop'
    for(SegmentSet::iterator
        leaf_iter  = leaf_segments_of_loop.begin();
        leaf_iter != leaf_segments_of_loop.end();
        leaf_iter++) {

      // Here we add to all 'leaf_segments_of_loop' the increment in the for
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'i++'. Afterwards, it will look like
      // 'int dd; i++;'.
      graph_[*leaf_iter].add_expression(for_stmt->get_increment());

      // Here we add to all 'leaf_segments_of_loop' the condition in the for
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'i<99'. Afterwards, it will look like
      // 'int dd; i++; i<99;'.
      graph_[*leaf_iter].add_expression(for_condition);

      // Here we add all the expressions of the 'loop_vertex' (int bb)
      // at the end of the last segment. So at this point the segment
      // 'int dd' has changed into 'int dd; i++; i<99; int bb;
      graph_[*leaf_iter].merge(graph_[loop_vertex]);
    }

    // Here we are adding all required expression to the
    // 'continue_segments'
    for(SegmentSet::iterator
        leaf_iter  = continue_segments.begin();
        leaf_iter != continue_segments.end();
        leaf_iter++) {

      // Here we add to all 'continue_segments' the increment in the for
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'i++'. Afterwards, it will look like
      // 'int dd; i++;'.
      graph_[*leaf_iter].add_expression(for_stmt->get_increment());

      // Here we add to all 'continue_segments' the condition in the for
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'i<99'. Afterwards, it will look like
      // 'int dd; i++; i<99;'.
      graph_[*leaf_iter].add_expression(for_condition);

      // Here we add all the expressions of the 'loop_vertex' (int bb)
      // at the end of the last segment. So at this point the segment
      // 'int dd' has changed into 'int dd; i++; i<99; int bb;
      graph_[*leaf_iter].merge(graph_[loop_vertex]);
    }

    // Here we are adding all required expression to the
    // 'continue_segments'
    for(SegmentSet::iterator
        current_segments_iter  = current_segments.begin();
        current_segments_iter != current_segments.end();
        current_segments_iter++) {

      // Here, we merge the current segments like 'int aa' with the first
      // segment of the loop. At the end we can delete the first node of the
      // loop because we are merge it here.
      graph_[*current_segments_iter].merge(graph_[loop_vertex]);

      // This covers the case that we delete the vertex 'loop_vertex' later,
      // however, the loop_vertex is also a leaving segment. So, we have to
      // consider the merged 'current_segment' as leaving segment as well.
      if(leaf_segments_of_loop.find(loop_vertex) !=
          leaf_segments_of_loop.end()) {

        new_current_segments.insert(*current_segments_iter);
      }
    }

    // In the following we add the 'leaf_segments_of_loop', the
    // 'current_segments', the 'break_segments', and the 'continue_segments'
    // to the 'new_current_segments'. These are the segments which will leave
    // the loop.
    new_current_segments.insert(leaf_segments_of_loop.begin(),
        leaf_segments_of_loop.end());
    new_current_segments.insert(current_segments.begin(),
        current_segments.end());
    new_current_segments.insert(break_segments.begin(),
        break_segments.end());
    new_current_segments.insert(continue_segments.begin(),
        continue_segments.end());

    // Here we remove the loop vertex from the graph because all the
    // statements already moved into the current segments.
    new_current_segments.erase(loop_vertex);
    boost::clear_vertex(loop_vertex, graph_);
    boost::remove_vertex(loop_vertex, graph_);

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(for_condition,
        new_current_segments);

    return new_current_segments;

  } else if (isSgDoWhileStmt(current_stmt)) { // do while loop

    // This example will be used to explain the following code.
    // The vertex descriptor 'loop_vertex' is a temporary segment graph node
    // for the loop which will be deleted after the loop processing. This node
    // includes all the statements until the boundary statements. In the given
    // example this would be 'int bb'.
    //
    // int aa;  <- current segments
    // do{
    //   int bb;  <- loop_vertex starts here
    //   wait();
    //   int cc;  <- boost::target(*out_edge_iter, graph_),
    //   wait();
    //   int dd;  <- leaf_segments_of_loop
    // } while(i < 99);
    // int ee;
    // wait();
    // int ff;

    SgDoWhileStmt *do_while_stmt = isSgDoWhileStmt(current_stmt);
    SgExpression *loop_condition =
      isSgExprStatement(do_while_stmt->get_condition())->get_expression();

    // Here we create a new segment for the loop body
    Segment loop_segment(NULL, false);
    VertexDescriptor loop_vertex = boost::add_vertex(loop_segment, graph_);
    SegmentSet leaf_segments_of_loop;
    leaf_segments_of_loop.insert(loop_vertex);

    // Analyze the loop body
    SegmentSet new_current_segments, break_segments, continue_segments;
    leaf_segments_of_loop = build_graph(do_while_stmt->get_body(),
                                leaf_segments_of_loop,
                                break_segments,
                                continue_segments,
                                duplicate_segments);

    // We are adding the edges among the individual segments
    insert_loop_edges(loop_vertex,
        current_segments,
        continue_segments,
        leaf_segments_of_loop);

    // Here we are adding all required expression to the
    // 'leaf_segments_of_the_loop'
    for(SegmentSet::iterator
        leaf_iter  = leaf_segments_of_loop.begin();
        leaf_iter != leaf_segments_of_loop.end();
        leaf_iter++) {

      // Here we add to all 'leaf_segments_of_loop' the condition in the for
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'i<99'. Afterwards, it will look like
      // 'int dd; i<99;'.
      graph_[*leaf_iter].add_expression(loop_condition);

      // Here we add all the expressions of the 'loop_vertex' (int bb)
      // at the end of the last segment. So at this point the segment
      // 'int dd' has changed into 'int dd; i<99; int bb;
      graph_[*leaf_iter].merge(graph_[loop_vertex]);
    }

    // Here we are adding all required expression to the
    // 'continue_segments'
    for(SegmentSet::iterator
        leaf_iter  = continue_segments.begin();
        leaf_iter != continue_segments.end();
        leaf_iter++) {

      // Here we add to all 'continue_segments' the condition in the for
      // loop. In the example above it would be adding in the segment 'int dd'
      // the condition expression 'i<99'. Afterwards, it will look like
      // 'int dd; i<99;'.
      graph_[*leaf_iter].add_expression(loop_condition);

      // Here we add all the expressions of the 'loop_vertex' (int bb)
      // at the end of the last segment. So at this point the segment
      // 'int dd' has changed into 'int dd; i<99; int bb;
      graph_[*leaf_iter].merge(graph_[loop_vertex]);
    }

    // Here we are adding all required expression to the
    // 'continue_segments'
    for(SegmentSet::iterator
        current_segments_iter  = current_segments.begin();
        current_segments_iter != current_segments.end();
        current_segments_iter++) {

      // Here, we merge the current segments like 'int aa' with the first
      // segment of the loop. At the end we can delete the first node of the
      // loop because we are merge it here.
      graph_[*current_segments_iter].merge(graph_[loop_vertex]);

      // This covers the case that we delete the vertex 'loop_vertex' later,
      // however, the loop_vertex is also a leaving segment. So, we have to
      // consider the merged 'current_segment' as leaving segment as well.
      if(leaf_segments_of_loop.find(loop_vertex) !=
          leaf_segments_of_loop.end()) {

        new_current_segments.insert(*current_segments_iter);
      }
    }

    // In the following we add the 'leaf_segments_of_loop', the
    // 'break_segments', and the 'continue_segments' to the
    // 'new_current_segments'. These are the segments which will leave the
    // loop.
    new_current_segments.insert(leaf_segments_of_loop.begin(),
                             leaf_segments_of_loop.end());
    new_current_segments.insert(break_segments.begin(),
                             break_segments.end());
    new_current_segments.insert(continue_segments.begin(),
                             continue_segments.end());

    // Here we remove the loop vertex from the graph because all the
    // statements already moved into the current segments.
    new_current_segments.erase(loop_vertex);
    boost::clear_vertex(loop_vertex, graph_);
    boost::remove_vertex(loop_vertex, graph_);

    // This case is only for the do while loop required. It describes
    // situation that in the entire loop body was no boundary statement.
    // So, we have to provide the incoming segments outgoing segments.
    if(new_current_segments.empty()) {
      new_current_segments = current_segments;
    }

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(loop_condition,
        new_current_segments);

    return new_current_segments;

  } else if(isSgBasicBlock(current_stmt)) { // basic block statement

    SgBasicBlock *bb_stmt = isSgBasicBlock(current_stmt);

    separate_variable_declaration_and_initializer_on_demand(bb_stmt,
        duplicate_segments);

    SgScopeStatement *enclosing_scope =
      SageInterface::getEnclosingScope(current_stmt);

    for(std::vector<SgStatement*>::const_iterator
        stmt_iter  = bb_stmt->get_statements().begin();
        stmt_iter != bb_stmt->get_statements().end();
        ++stmt_iter) {

      // here we decompose expression which include a function call that
      // has an indirect boundary statement. Also we will decompose member
      // function calls with boundary statements
      SgBasicBlock *bb = NULL;
      if(isSgExprStatement(*stmt_iter)) {
        SgExpression *expr = isSgExprStatement(*stmt_iter)->get_expression();
        decompose_expression_with_boundary_calls(expr, bb, duplicate_segments);
      }

      SegmentSet temp_break_segments, temp_continue_segments;
      current_segments = build_graph(*stmt_iter,
                             current_segments,
                             temp_break_segments,
                             temp_continue_segments,
                             duplicate_segments);

      break_segments.insert(temp_break_segments.begin(),
                         temp_break_segments.end());
      continue_segments.insert(temp_continue_segments.begin(),
                            temp_continue_segments.end());

      if(current_segments.empty()) {
        break;
      }

      if(isSgBreakStmt(*stmt_iter)) {
        break;
      }

      if(isSgContinueStmt(*stmt_iter)) {
        break;
      }
    }

  } else if(isSgSwitchStatement(current_stmt)) { // switch statement

    SgSwitchStatement *switch_stmt = isSgSwitchStatement(current_stmt);

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    SgExprStatement *expr_stmt
      = isSgExprStatement(switch_stmt->get_item_selector());
    mark_not_conflict_free_func_calls_in_segment(
        expr_stmt->get_expression(), current_segments);

    SegmentSet break_segments;
    current_segments = build_graph(switch_stmt->get_item_selector(),
                            current_segments,
                            break_segments,
                            continue_segments,
                            duplicate_segments);

    std::vector<SgStatement*> switch_cases
      = SageInterface::getSwitchCases(switch_stmt);

    SegmentSet leaving_segments;
    SegmentSet start_segments(current_segments);
    SegmentSet result;

    bool default_cases_exists = false;

    // Strategy for analyzing switch statements
    //
    // Let's discuss this example:
    // switch(var)                  <- Segment 0
    // {
    //   case 1: int aa;
    //           if(condition) {
    //             int bb;
    //             wait();          <- creates segment 1
    //             int cc;
    //           } else {
    //             int dd;
    //             wait();          <- creates segment 2
    //             int ee;
    //           }
    //   case 2: int ff;
    //   case 3: ...
    // }
    //
    // In the given example, the case 1 causes two new segments because of the
    // if-else in combination with both waits. We have to continue with both
    // segments for all following segments (case 2, case 3, ...)until we hit a
    // break condition.
    // For each case individually, we have to start we the 'current_segments'

    // here we analyze each segment
    for(std::vector<SgStatement*>::iterator
        case_iter  = switch_cases.begin();
        case_iter != switch_cases.end();
        case_iter++) {

      SgCaseOptionStmt *switch_case = isSgCaseOptionStmt(*case_iter);
      SgDefaultOptionStmt *default_case = isSgDefaultOptionStmt(*case_iter);
      SgBasicBlock *bb = NULL;

      if(switch_case)
        bb = isSgBasicBlock(switch_case->get_body());

      if(default_case) {
        bb = isSgBasicBlock(default_case->get_body());
        default_cases_exists = true;
      }

      SegmentSet break_segments;
      result = build_graph(bb,
                   start_segments,
                   break_segments,
                   continue_segments,
                   duplicate_segments);

      leaving_segments.insert(break_segments.begin(), break_segments.end());

      // This case describes the situation that we are iterating the last
      // switch case or the default case. So, these segments are definitely
      // are leaving segments.
      if(case_iter+1 == switch_cases.end())
        leaving_segments.insert(result.begin(), result.end());

      // The current switch case has no boundary statements. So, the starting
      // segment set 'start_segments' and the resulting segment set 'result'
      // are the same. We can skip here.
      if(result == start_segments)
        continue;

      // The result set is empty when we hit a break statement
      // At this point, no following cases must be analyzed.
      // The break segments will be added later to the leaving segments
      if(result.empty()) {
        continue;
      }

      // At this position, in the switch case a boundary statement was hit.
      // This can be compared with 'case 1' in the example about. We started
      // with segment 0 and now we continue with segment 1 and 2. So, we will
      // iterate over the following cases
      for(std::vector<SgStatement*>::iterator
          following_segment_iter  = case_iter + 1;
          following_segment_iter != switch_cases.end();
          following_segment_iter++) {

        switch_case = isSgCaseOptionStmt(*following_segment_iter);
        default_case = isSgDefaultOptionStmt(*following_segment_iter);
        bb = NULL;

        if(switch_case)
          bb = isSgBasicBlock(switch_case->get_body());

        if(default_case)
          bb = isSgBasicBlock(default_case->get_body());

        // Like before we have to process all the segments
        SegmentSet break_segments;
        result = build_graph(bb,
                     result,
                     break_segments,
                     continue_segments,
                     duplicate_segments);

        leaving_segments.insert(break_segments.begin(), break_segments.end());

        // The result set is empty when we hit a break statement
        // At this point, no following cases must be analyzed.
        // The break segments will be added later to the leaving segments
        if(result.empty()) {
          break;
        }
      }

      // finally, we add all segments to the 'leaving_segments'
      leaving_segments.insert(result.begin(), result.end());
    }

    leaving_segments.insert(break_segments.begin(), break_segments.end());
    break_segments.clear();

    if(!default_cases_exists) {
      leaving_segments.insert(current_segments.begin(), current_segments.end());
    }

    return leaving_segments;

  } else if(isSgExprStatement(current_stmt)) { // expression statement

    // At this point we know that the expression can have at most one
    // function call which has a boundary statement.
    SgExpression *expr = isSgExprStatement(current_stmt)->get_expression();

    // Here, we check if there is any function call in the expression which
    // is annotated as not-conflict-free
    mark_not_conflict_free_func_calls_in_segment(expr, current_segments);

    // func_call_exp has the node of the function call with the boundary
    SgFunctionCallExp *func_call_exp
      = has_function_call_with_boundary(expr, duplicate_segments);

    if(func_call_exp) {

      FunctionAnnotationAttributes *function_annotation
        = get_function_call_annotation(func_call_exp);

      if(function_annotation) {

        // We know if this case, that there must be a boundary.
        // So, we don't have to double check with has_boundary()
        CachedFunctionAstAttributes *cached_graph
          = create_cached_function_attribute_for_annotated_function(
                func_call_exp,
                function_annotation->wait_type_,
                current_segments,
                function_annotation->is_conflict_free_);

        return cached_graph->leaving_segments_;
      }

      SgFunctionDeclaration *func_decl
        = tools::get_first_non_defining_function_declaration(func_call_exp);

      if(func_decl == NULL) {

        add_expression_to_segment(current_segments, func_call_exp);

        return current_segments;
      }

      func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());
      SgFunctionDefinition *func_def = func_decl->get_definition();

      tools::AstAttributeColor *call_hierarchy_color =
        dynamic_cast<tools::AstAttributeColor*>(
            func_def->getAttribute("function_call_hierarchy"));

      // In this case, the function was already discovered and the function
      // is part of the current function call hierarchy. It's a recursive
      // function call.
      if(call_hierarchy_color->call_hierarchy_ ==
          tools::AstAttributeColor::IN_CALL_HIERARCHY) {

        // Strategy:
        // If we have a recursive function call, we will put the function
        // call in a separated segment. This segment contains only this
        // function call and will be deleted in a later stage.
        //
        // +------------------------------------+
        // |  Statements before recursive call  |
        // +-----------------+------------------+
        //                   |
        //                   v
        // +------------------------------------+
        // |       Recursive function call      |
        // +-----------------+------------------+
        //                   |
        //                   v
        // +------------------------------------+
        // |        Follow up statements        |
        // +------------------------------------+
        //
        // The recursive segment will stored in the recursive vector and
        // processed later.

        // This will be the temporally segment for function call
        Segment call_expr_segment(NULL, false);
        VertexDescriptor func_call_vertex =
          boost::add_vertex(call_expr_segment, graph_);
        recursive_function_calls_.insert(func_call_vertex);
        SegmentSet call_expr_segment_set;
        call_expr_segment_set.insert(func_call_vertex);

        add_expression_to_segment(call_expr_segment_set, func_call_exp);

        for(SegmentSet::iterator
            curr_seg_iter  = current_segments.begin();
            curr_seg_iter != current_segments.end();
            curr_seg_iter++) {

          if(!boost::edge(*curr_seg_iter,
                         func_call_vertex,
                         graph_).second) {

            boost::add_edge(*curr_seg_iter, func_call_vertex, graph_);
          }
        }

        // This is the follow up segment of the temporally function call
        // segment
        Segment leaving_segment(NULL, false);
        VertexDescriptor leaving_vertex =
          boost::add_vertex(leaving_segment, graph_);
        SegmentSet leaving_segment_set;
        leaving_segment_set.insert(leaving_vertex);

        if(!boost::edge(func_call_vertex,
                       leaving_vertex,
                       graph_).second) {

          boost::add_edge(func_call_vertex, leaving_vertex, graph_);
        }

        call_hierarchy_color->call_hierarchy_ =
          tools::AstAttributeColor::NOT_IN_CALL_HIERARCHY;

        return leaving_segment_set;

      } else {
        // In this case, the function was already discovered, however, the
        // it wasn't in the function call hierarchy up to now. We mark it
        // as part of the function call hierarchy now.
        call_hierarchy_color->call_hierarchy_ =
          tools::AstAttributeColor::IN_CALL_HIERARCHY;
        func_def->setAttribute("function_call_hierarchy",
            call_hierarchy_color);
      }

      add_expression_to_segment(current_segments, current_stmt);

      // here we are adding and merging all functions which are called
      // as an argument
      std::vector<SgNode*> func_calls =
        NodeQuery::querySubTree(func_call_exp->get_args(), V_SgFunctionCallExp);

      for(std::vector<SgNode*>::iterator
          func_call_iter  = func_calls.begin();
          func_call_iter != func_calls.end();
          func_call_iter++) {

        SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(*func_call_iter);
        SgFunctionDeclaration *func_decl
          = tools::get_first_non_defining_function_declaration(func_call_exp);

        if(func_decl) {

          func_decl
            = isSgFunctionDeclaration(func_decl->get_definingDeclaration());
          SgFunctionDefinition *func_def = func_decl->get_definition();

          if(func_decl != NULL) {

            if(follow_function_call(func_decl)) {

              CachedFunctionAstAttributes *precached =
                dynamic_cast<CachedFunctionAstAttributes*>(
                    func_def->getAttribute("precached_info"));

              for(SegmentSet::iterator
                  iter_current_segment  = current_segments.begin();
                  iter_current_segment != current_segments.end();
                  iter_current_segment++) {

                // Here, we merge the beginning segment of the function into the
                // current flow
                graph_[*iter_current_segment].
                  merge(graph_[precached->starting_vertex_]);
              }
            }
          } else {
            // THIS is a virtual function
            return current_segments;
          }
        }
      }

      // cloning of segments for channels
      if(tools::function_call_belongs_to_a_port(func_call_exp)) {

        CachedFunctionAstAttributes *precached_function =
          dynamic_cast<CachedFunctionAstAttributes*>(
              func_def->getAttribute("precached_info"));

        std::pair<risc::sg::VertexDescriptor,
                  risc::sg::SegmentGraph::SegmentSet> result
          = risc::sg::clone_graph(
                precached_function->starting_vertex_,
                precached_function->leaving_segments_,
                *this,
                true);

        // The first segment of the cloned channel will be merged.
        // So, we put it into the set of segments for deletion
        precached_function_segments_.insert(result.first);

        // This will be used later during the instrumentation
        PortCallAttribute *port_call_attribute =
          dynamic_cast<PortCallAttribute*>(
              func_call_exp->getAttribute("port_call"));

        // it is possible that this is already set
        // example if (test) wait() else wait() portcall
        // port call would be in both segments
        if(port_call_attribute == NULL) {

          std::list<int> reachable_ids
            = get_all_reachable_segments(*this, graph_[result.first].id_);

          PortCallAttribute *port_call_attribute = new PortCallAttribute();

          func_call_exp->setAttribute("port_call", port_call_attribute);
          port_calls_.insert(func_call_exp);


          // we skip the fist node because that will be merged and deleted later
          std::list<int>::iterator iter  = reachable_ids.begin();
          iter++;

          for(int j = 0; iter != reachable_ids.end(); iter++, j++ ) {

            port_call_attribute->segment_ids_.push_back(
                &graph_[id_to_vertex_descriptor(*iter)]);

            graph_[id_to_vertex_descriptor(*iter)].wait_offset_ = j;
          }
        }

        // We integrate the pre analyzed function into the current segments.
        // The first segment will be merged
        for(SegmentSet::iterator
            current_segment  = current_segments.begin();
            current_segment != current_segments.end();
            current_segment++) {

          // Here, we merge the beginning segment of the function into the
          // current flow
          graph_[*current_segment].merge(graph_[result.first]);

          // In the following, we connect the successors of the beginning
          // segments of the function with the current segments
          Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
          boost::tie(out_edge_begin, out_edge_end) =
            boost::out_edges(result.first, graph_);

          for(out_edge_iter  = out_edge_begin;
              out_edge_iter != out_edge_end;
              out_edge_iter++) {

            if(!boost::edge(*current_segment,
                           boost::target(*out_edge_iter, graph_),
                           graph_).second) {

              boost::add_edge(*current_segment,
                       boost::target(*out_edge_iter, graph_),
                       graph_);
            }
          }
        }

        call_hierarchy_color->call_hierarchy_ =
          tools::AstAttributeColor::NOT_IN_CALL_HIERARCHY;

        // We have to check if the starting vertex is a leaving vertex as well.
        // If that is the case, then we have to consider all current segments as
        // leaving segments as well.
        if(precached_function->starting_segment_is_leaving_segment()) {

          SegmentSet return_segments(result.second);

          return_segments.erase(result.first);

          return_segments.insert(current_segments.begin(),
                              current_segments.end());

          std::vector<VertexDescriptor> reachable
            = get_all_reachable_segments(*this,
                precached_function->starting_vertex_);

          for(std::vector<VertexDescriptor>::iterator
              iter  = reachable.begin();
              iter != reachable.end();
              iter++) {
            if(graph_[*iter].id_ != graph_[result.first].id_) {
              precached_function_segments_.insert(*iter);
            }
          }

          return return_segments;

        } else {

          std::vector<VertexDescriptor> reachable
            = get_all_reachable_segments(*this,
                precached_function->starting_vertex_);

          precached_function_segments_.insert(
              reachable.begin(), reachable.end());
          precached_function_segments_.insert(result.first);

          return result.second;
        }

      } else {

        CachedFunctionAstAttributes *precached_function =
          dynamic_cast<CachedFunctionAstAttributes*>(
              func_def->getAttribute("precached_info"));

        // The function was already discovered. Now, we integrate the pre
        // analyzed function into the current part of the segment graph
        for(SegmentSet::iterator
            current_segment  = current_segments.begin();
            current_segment != current_segments.end();
            current_segment++) {

          // Here, we merge the beginning segment of the function into the
          // current flow
          graph_[*current_segment].merge(
              graph_[precached_function->starting_vertex_]);

          // In the following, we connect the successors of the beginning
          // segments of the function with the current segments
          Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
          boost::tie(out_edge_begin, out_edge_end) =
            boost::out_edges(precached_function->starting_vertex_, graph_);

          VertexDescriptor temp = precached_function->starting_vertex_;
          for(out_edge_iter  = out_edge_begin;
              out_edge_iter != out_edge_end;
              out_edge_iter++) {

            if(!boost::edge(*current_segment,
                           boost::target(*out_edge_iter, graph_),
                           graph_).second) {

              boost::add_edge(*current_segment,
                       boost::target(*out_edge_iter, graph_),
                       graph_);
            }
          }
        }

        call_hierarchy_color->call_hierarchy_ =
          tools::AstAttributeColor::NOT_IN_CALL_HIERARCHY;

        // We have to check if the starting vertex is a leaving vertex as well.
        // If that is the case, then we have to consider all current segments as
        // leaving segments as well.
        if(precached_function->starting_segment_is_leaving_segment()) {

          SegmentSet return_segments(precached_function->leaving_segments_);
          return_segments.erase(precached_function->starting_vertex_);
          return_segments.insert(current_segments.begin(),
                              current_segments.end());
          return return_segments;

        } else {

          return precached_function->leaving_segments_;
        }
      }

    } else { // This expression is boundary function call free

      add_expression_to_segment(current_segments, expr);

      std::vector<SgNode*> func_calls =
        NodeQuery::querySubTree(expr, V_SgFunctionCallExp);

      // So, all function calls have no boundary statements
      // Nevertheless, we have to merge all calls into the current segment
      for(std::vector<SgNode*>::iterator
          func_call_iter  = func_calls.begin();
          func_call_iter != func_calls.end();
          func_call_iter++) {

        SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(*func_call_iter);
        SgFunctionDeclaration *func_decl
          = tools::get_first_non_defining_function_declaration(func_call_exp);

        if(func_decl) {

          func_decl =
            isSgFunctionDeclaration(func_decl->get_definingDeclaration());

          if(func_decl) {

            SgFunctionDefinition *func_def = func_decl->get_definition();

            if(func_decl != NULL) {
              CachedFunctionAstAttributes *precached =
                dynamic_cast<CachedFunctionAstAttributes*>(
                    func_def->getAttribute("precached_info"));

              if(precached) {
                for(SegmentSet::iterator
                    iter_current_segment  = current_segments.begin();
                    iter_current_segment != current_segments.end();
                    iter_current_segment++) {

                  // Here, we merge the beginning segment of the function into
                  // the current flow
                  graph_[*iter_current_segment].
                    merge(graph_[precached->starting_vertex_]);
                }
              }
            }
          }
        }
      }
    }

    return current_segments;

  } else if(isSgVariableDeclaration(current_stmt)) { // variable declaration

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(current_stmt);
    add_expression_to_segment(current_segments, var_decl);

    // here we are linking the references
    SgVariableSymbol *symbol
      = isSgVariableSymbol(var_decl->get_decl_item(0)
          ->search_for_symbol_from_symbol_table());

    if(isSgReferenceType(symbol->get_type())) {

      SgAssignInitializer *assign_init
        = isSgAssignInitializer(symbol->get_declaration()->get_initializer());
      SgVariableSymbol *initializer_symbol
        = tools::get_most_outer_var(assign_init->get_operand_i());

      // If the function call considers has only literals there will be no
      // symbol
      if(initializer_symbol) {
        LocalReferenceAttribute *local_mapping = new LocalReferenceAttribute();
        symbol->setAttribute("local_reference", local_mapping);

        if(tools::is_local_variable(initializer_symbol)) {

          if(isSgReferenceType(initializer_symbol->get_type())) {

            local_mapping->mapped_variable_
              = dynamic_cast<LocalReferenceAttribute*>(
                  initializer_symbol->getAttribute("local_reference"))
                                        ->mapped_variable_;
          } else {

            local_mapping->mapped_variable_ = initializer_symbol;
          }
        } else {

          local_mapping->mapped_variable_ = initializer_symbol;
        }
      }
    }

    // here we consider situations like 'int a = foo()' or
    // 'int array[] = {foo()};' where 'foo()' doesn't include any direct
    // or indirect boundary statement
    SgInitializedName *initialized_name = var_decl->get_decl_item(0);

    if(initialized_name->get_initptr()) {

      std::vector<SgNode*> func_calls;

      if(isSgAssignInitializer(initialized_name->get_initptr())) {
        func_calls = NodeQuery::querySubTree(
                                    isSgAssignInitializer(
                                        initialized_name->get_initptr())->
                                        get_operand(),
                                    V_SgFunctionCallExp);
      }

      if(isSgAggregateInitializer(initialized_name->get_initptr())) {

        func_calls = NodeQuery::querySubTree(
            isSgAggregateInitializer(
                initialized_name->get_initptr())->get_initializers(),
            V_SgFunctionCallExp);
      }

      for(std::vector<SgNode*>::iterator
          iter_func_calls  = func_calls.begin();
          iter_func_calls != func_calls.end();
          iter_func_calls++) {

        SgFunctionCallExp *func_call = isSgFunctionCallExp(*iter_func_calls);

        // Here, we check if there is any function call in the expression which
        // is annotated as not-conflict-free
        mark_not_conflict_free_func_calls_in_segment(func_call,
            current_segments);

        SgFunctionDeclaration *func_decl
          = tools::get_first_non_defining_function_declaration(func_call);

        // In this case we are facing a function which is defined in a
        // library. So, we skip the analysis of this one
        if(func_decl == NULL) {
          continue;
        }

        if(follow_function_call(func_decl)) {

          func_decl =
            isSgFunctionDeclaration(func_decl->get_definingDeclaration());

          if(func_decl == NULL) {
            continue;
          }

          SgFunctionDefinition *func_def = func_decl->get_definition();

          CachedFunctionAstAttributes *precached_function =
            dynamic_cast<CachedFunctionAstAttributes*>(
                func_def->getAttribute("precached_info"));

          for(SegmentSet::iterator
              iter_current_segments  = current_segments.begin();
              iter_current_segments != current_segments.end();
              iter_current_segments++) {

            graph_[*iter_current_segments].merge(
                graph_[precached_function->starting_vertex_]);
          }
        }
      }
    }

    return current_segments;

  } else if(isSgBreakStmt(current_stmt)) { // Break Statement

    SgBreakStmt *break_stmt = isSgBreakStmt(current_stmt);
    break_segments.insert(current_segments.begin(), current_segments.end());
    add_expression_to_segment(current_segments, break_stmt );
    current_segments.clear();
    return current_segments;

  } else if(isSgContinueStmt(current_stmt)) { // Continue Statement

    SgContinueStmt *continue_stmt = isSgContinueStmt(current_stmt);
    continue_segments.insert(current_segments.begin(), current_segments.end());
    add_expression_to_segment(current_segments, continue_stmt );
    current_segments.clear();
    return current_segments;

  } else if(isSgDeclarationStatement(current_stmt)) { // some other declaration
                                                      // statement

    SgDeclarationStatement *decl = isSgDeclarationStatement(current_stmt);
    return current_segments;

  } else if (isSgReturnStmt(current_stmt)) {

    SgReturnStmt *return_stmt = isSgReturnStmt(current_stmt);
    SgExpression *expr = return_stmt->get_expression();
    add_expression_to_segment(current_segments, expr);

    SgFunctionDefinition *func_def =
      SageInterface::getEnclosingFunctionDefinition(current_stmt);

    CachedFunctionAstAttributes *precached_function =
      dynamic_cast<CachedFunctionAstAttributes*>(
          func_def->getAttribute("precached_info"));

    // here we are adding the leave segments through a return statement
    // the thread function itself has not
    if(precached_function) {
      precached_function->leaving_segments_.insert(current_segments.begin(),
                                                current_segments.end());

      current_segments.clear();
    }

    return current_segments;
  }

  return current_segments;
}

bool
risc::sg::SegmentGraph::is_boundary_stmt(
    SgStatement const * const current_stmt)
{
  // here we check if we call given function
  // TODO: How to handle virtual interfaces?
  SgExprStatement const * const expr_stmt = isSgExprStatement(current_stmt);
  if(expr_stmt) {

    SgFunctionCallExp const * const func_call_exp
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call_exp) {

      for(FunctionCallBoundarySet::const_iterator
          func_def_iter  = function_boundaries_.begin();
          func_def_iter != function_boundaries_.end();
          func_def_iter++) {

        if(func_call_exp->getAssociatedFunctionSymbol() ==
            (*func_def_iter)->get_declaration()->
              search_for_symbol_from_symbol_table()) {
          return true;
        }
      }
    }
  }

  for(CCxxKeywordsBoundarySet::const_iterator
      keyword_iter  = keyword_boundaries_.begin();
      keyword_iter != keyword_boundaries_.end();
      keyword_iter++) {

    if((*keyword_iter) == current_stmt->variantT()) {
      return true;
    }
  }

  return false;
}

void
risc::sg::SegmentGraph::insert_loop_edges(VertexDescriptor &loop_vertex,
    SegmentSet &current_segments, SegmentSet &continue_segments,
    SegmentSet &leaf_segments_of_loop)
{
  // This example will be used to explain the following code. The handling for
  // the different loop types (for, while, and do while) is identical.
  // The vertex descriptor 'loop_vertex' is a temporary segment graph node
  // for the loop which will be deleted after the loop processing. This node
  // includes all the statements until the boundary statements. In the given
  // example this would be 'int bb'.
  //
  // int aa;  <- current segments
  // while(condition) {
  //   int bb;  <- loop_vertex starts here
  //   wait();
  //   int cc;  <- boost::target(*out_edge_iter, graph_),
  //   wait();
  //   int dd;  <- leaf_segments_of_loop
  // }
  // int ee;
  // wait();
  // int ff;

  Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
  boost::tie(out_edge_begin, out_edge_end) =
    boost::out_edges(loop_vertex, graph_);

  // We are iterating over all outgoing edges of the temporarily node
  // 'loop_vertex'
  for(out_edge_iter  = out_edge_begin;
      out_edge_iter != out_edge_end;
      out_edge_iter++) {

    // Here, we are connecting the 'current_segments' with all successors of
    // the 'loop_vertex' segment. We are doing this because the 'loop_vertex'
    // will be merged into the 'current_segments' later and deleted
    // afterwards.
    for(SegmentSet::iterator
        current_segments_iter  = current_segments.begin();
        current_segments_iter != current_segments.end();
        current_segments_iter++) {

      if(!boost::edge(*current_segments_iter,
                     boost::target(*out_edge_iter, graph_),
                     graph_).second) {

        boost::add_edge(*current_segments_iter,
            boost::target(*out_edge_iter, graph_), graph_);
      }
    }

    // 'continue_segments' are the segments which contain a continue. So,
    // the control flow will go back to the loop condition and than (maybe)
    // take another loop iteration.
    // Here, we are connecting the 'continue_segments' with all successors of
    // the 'loop_vertex' segment. We are doing this because the 'loop_vertex'
    // will be merged into the 'current_segments' later and deleted
    // afterwards.
    for(SegmentSet::iterator
        continue_segments_iter  = continue_segments.begin();
        continue_segments_iter != continue_segments.end();
        continue_segments_iter++) {

      if(!boost::edge(*continue_segments_iter,
            boost::target(*out_edge_iter, graph_), graph_).second) {

        boost::add_edge(*continue_segments_iter,
            boost::target(*out_edge_iter, graph_), graph_);
      }
    }

    // 'leaf_segments_of_loop' are the segments which finish the loop body
    // without facing a continue or a break. So, the control flow will go back
    // to the loop condition and than (maybe) take another loop iteration.
    // Here, we are connecting the 'leaf_segments_of_loop' with all successors
    // of the 'loop_vertex' segment. We are doing this because the
    // 'loop_vertex' will be merged into the 'current_segments' later and
    // deleted afterwards.
    for(SegmentSet::iterator
        leaf_segments_iter  = leaf_segments_of_loop.begin();
        leaf_segments_iter != leaf_segments_of_loop.end();
        leaf_segments_iter++) {

      if(!boost::edge(*leaf_segments_iter,
            boost::target(*out_edge_iter, graph_), graph_).second) {

        boost::add_edge(*leaf_segments_iter,
            boost::target(*out_edge_iter, graph_), graph_);
      }
    }
  }
}

void
risc::sg::SegmentGraph::read_write_analysis_of_segments()
{
  // this will be done conflict table class
  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator vertex_iter;
  std::pair<vertex_iter, vertex_iter> vp;
  for(vp = boost::vertices(graph_); vp.first != vp.second; ++vp.first) {

    graph_[*vp.first].analyze_segment();
  }
}

SgFunctionCallExp*
risc::sg::SegmentGraph::has_function_call_with_boundary(SgExpression *expr,
    bool duplicate_segments)
{
  // At this point, an expression can only have at max one function call
  // which has a boundary.
  // Expressions with multiple boundary function calls are decomposed
  // while iterating over the associated basic block

  std::vector<SgNode*> func_calls
    = NodeQuery::querySubTree(expr, V_SgFunctionCallExp);

  for(std::vector<SgNode*>::iterator
      iter  = func_calls.begin();
      iter != func_calls.end();
      iter++) {

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(*iter);

    // Here we are checking, if the function has some annotation.
    // If so, we can return and we don't have to analyze the function
    FunctionAnnotationAttributes *annotation
      = get_function_call_annotation(func_call_exp);
    if(annotation) {
      if(annotation->has_boundary()) {
        return func_call_exp;
      }
      continue;
    }

    // Okay, the function has a definition and no pragma statement.
    // So, we have to analyze the function
    SgFunctionDeclaration *func_decl
      = tools::get_first_non_defining_function_declaration(func_call_exp);

    if(func_decl) {

      func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());

      if(func_decl) {

        if(follow_function_call(func_decl)) {

          SgFunctionDefinition *func_def = func_decl->get_definition();
          CachedFunctionAstAttributes *precached_function =
            dynamic_cast<CachedFunctionAstAttributes*>(
                func_def->getAttribute("precached_info"));

          if(!precached_function) {

            // Here, we are adding the arguments which are passed by reference
            // to the corresponding parameter in the function declaration
            // We need this information later for the conflict analysis
            add_mapped_symbol_to_reference(func_call_exp, func_def);

            precached_function
              = build_segment_graph_for_function(func_def, duplicate_segments);

          } else {

            // This function is already analyzed and we don't want to analyze
            // it twice (decomposition etc.). So, we are following the function
            // call and we are adding arguments passed by reference to the
            // definition. Afterwards, we get all the function call expressions
            // and forward references recursively.
            forward_reference_function_parameters(func_call_exp);
          }

          if(precached_function->has_boundary()) {
            return func_call_exp;
          }
        }
      }
    }
  }

  // There was no boundary function call
  return NULL;
}

bool
risc::sg::SegmentGraph::should_decompose_function(
    SgExpression *expr,
    bool duplicate_segments)
{
  // In the beginning we are checking if an assigment happens.
  // This can be via the assignment operator or an overloaded function '='
  SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expr);
  SgVariableSymbol *var_symbol = NULL;

  bool assignment = false;

  // This one is needed for overloaded assignment operators
  if(func_call_exp) {

    SgDotExp *dot = isSgDotExp(func_call_exp->get_function());

    if(dot) {

      SgMemberFunctionRefExp *func_ref
        = isSgMemberFunctionRefExp(dot->get_rhs_operand());

      if(func_ref) {

        if(func_ref->unparseToString() == "=") {
          assignment = true;

          SgArrowExp *arrow_exp = isSgArrowExp(dot->get_lhs_operand());

          if(arrow_exp) {
            var_symbol
              = tools::get_most_outer_var(arrow_exp->get_rhs_operand());

          } else {
            var_symbol = tools::get_most_outer_var(dot->get_lhs_operand());
          }

          func_call_exp
            = isSgFunctionCallExp(
                  func_call_exp->get_args()->get_expressions().at(0));
        }
      }
    }
  }

  if(assignment == false) {

    SgAssignOp *assign_op = isSgAssignOp(expr);
    if(assign_op) {

      var_symbol = tools::get_most_outer_var(assign_op->get_lhs_operand());

      if(var_symbol == NULL) {
        return true;
      }
      func_call_exp = isSgFunctionCallExp(assign_op->get_rhs_operand());
      assignment = true;
    }
  }

  // In this function we check if a function has the structure of
  // static_var/member_var = func_with_boundary(). If so, we have to decompose
  // it.
  // We don't want to decompose something like local_var = func_with_boundary().
  // This would end in a recursive decomposition
  if(assignment) {

    bool is_local_none_static_var = false;
    if(tools::is_local_variable(var_symbol)) {
      if(!tools::is_static_variable(var_symbol)) {
        is_local_none_static_var = true;
      }
    }

    if(func_call_exp && is_local_none_static_var) {

      std::vector<SgNode*> func_calls =
        NodeQuery::querySubTree(func_call_exp->get_args(), V_SgFunctionCallExp);

      for(std::vector<SgNode*>::iterator
          iter  = func_calls.begin();
          iter != func_calls.end();
          iter++) {

        SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(*iter);

        FunctionAnnotationAttributes *annotation
          = get_function_call_annotation(func_call_exp);

        if(annotation) {
          if(annotation->has_boundary()) {
            return true;
          }
        } else {

          SgFunctionDeclaration *func_decl
            = tools::get_first_non_defining_function_declaration(func_call_exp);

          if(func_decl) {
            func_decl = isSgFunctionDeclaration(
                            func_decl->get_definingDeclaration());

            SgFunctionDefinition *func_def = func_decl->get_definition();

            CachedFunctionAstAttributes *precached_function =
              dynamic_cast<CachedFunctionAstAttributes*>(
                  func_def->getAttribute("precached_info"));

            if(precached_function->has_boundary()) {
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  return true;
}

SgFunctionDeclaration*
risc::sg::SegmentGraph::contains_function_call_expression(
    SgExpression *expr)
{
  if(isSgAssignOp(expr)) {
    SgAssignOp *assign_op = isSgAssignOp(expr);
    return contains_function_call_expression(assign_op->get_rhs_operand());
  }

  if(isSgDotExp(expr)) {
    SgDotExp *dot_exp = isSgDotExp(expr);
    return contains_function_call_expression(dot_exp->get_rhs_operand());
  }

  if(isSgArrowExp(expr)) {
    SgArrowExp *arrow_exp = isSgArrowExp(expr);
    return contains_function_call_expression(arrow_exp->get_rhs_operand());
  }

  if(isSgFunctionCallExp(expr)) {
    return tools::get_first_non_defining_function_declaration(
                      isSgFunctionCallExp(expr));
  }

  if(isSgMemberFunctionRefExp(expr)) {
    return isSgMemberFunctionRefExp(expr)->
      getAssociatedMemberFunctionDeclaration();
  }

  return NULL;
}

void
risc::sg::SegmentGraph::handle_recursive_calls()
{
  // Here, we are handing the recursive function calls which we stored in the
  // recursive_function_calls_ vector.
  //
  // Function call handling:
  // The recursive segment will be merged with the statements of the beginning
  // segment of the called function. Afterwards, there will be an edge to the
  // successors of the beginning segment of the called function.
  //
  // Function return handling:
  // The successor segment 's' of the recursive function call segment will be
  // merged into the leaving segments of the recursive called functions.
  // Afterwards, there will be edges from the leaving segments of the recursive
  // called function to the successors of 's'

  // Iterating over all recursive calls
  for(std::set<VertexDescriptor>::iterator
      recursive_iter  = recursive_function_calls_.begin();
      recursive_iter != recursive_function_calls_.end();
      recursive_iter++) {

    SgFunctionCallExp *func_call =
      isSgFunctionCallExp(*graph_[*recursive_iter].expressions_.begin());

    SgFunctionDeclaration *func_decl
      = tools::get_first_non_defining_function_declaration(func_call);

    if(func_decl) {

      func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());
      SgFunctionDefinition *func_def = func_decl->get_definition();

      CachedFunctionAstAttributes *precached_function =
        dynamic_cast<CachedFunctionAstAttributes*>(
            func_def->getAttribute("precached_info"));

      //
      // +---------------------------------------+
      // |  Statements before recursive call (A) |
      // +-------------------+-------------------+
      //                     |
      //                     v
      // +---------------------------------------+
      // |       Recursive function call     (B) |
      // +-------------------+-------------------+
      //                     |
      //                     v
      // +---------------------------------------+
      // |        Follow up statements       (C) |
      // +---------------------------------------+
      //

      // These are the edges A->B
      Graph::in_edge_iterator in_edge_begin, in_edge_end, in_edge_iter;
      boost::tie(in_edge_begin, in_edge_end) =
        boost::in_edges(*recursive_iter, graph_);

      // These edges belong to the called function. The edges are between
      // the starting segment and the following up segment of the function.
      Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
      boost::tie(out_edge_begin, out_edge_end) =
        boost::out_edges(precached_function->starting_vertex_, graph_);

      // Here, we are inserting edges for the function call as described above
      // We are iterating over the edges A->B
      for(in_edge_iter  = in_edge_begin;
          in_edge_iter != in_edge_end;
          in_edge_iter++) {

        // here we are merging B into A
        graph_[boost::source(*in_edge_iter, graph_)].merge(
            graph_[*recursive_iter]);

        // Iterating over the edges of the called function
        for(out_edge_iter  = out_edge_begin;
            out_edge_iter != out_edge_end;
            out_edge_iter++) {

          // here we are adding the edge from A to the called function
          if(!boost::edge(boost::source(*in_edge_iter, graph_),
                         boost::target(*out_edge_iter, graph_), graph_)
                      .second) {

            boost::add_edge(boost::source(*in_edge_iter, graph_),
                       boost::target(*out_edge_iter, graph_),
                       graph_);
          }
        }

        // merging the starting vertex of the called function into A
        graph_[boost::source(*in_edge_iter, graph_)].merge(
            graph_[precached_function->starting_vertex_]);

        // here, we have to check if starting segment is also a leaving segment.
        // If this is the case, we have to do the following steps
        if(precached_function->starting_segment_is_leaving_segment()) {

          // we have to merge C into A
          Graph::out_edge_iterator out_edge_begin_func_seg,
            out_edge_end_func_seg, out_edge_iter_func_seg;

          boost::tie(out_edge_begin_func_seg, out_edge_end_func_seg) =
            boost::out_edges(*recursive_iter, graph_);

          // iterating over the B->C edges
          for(out_edge_iter_func_seg  = out_edge_begin_func_seg;
              out_edge_iter_func_seg != out_edge_end_func_seg;
              out_edge_iter_func_seg++) {

            // merging the C into A
            graph_[boost::source(*in_edge_iter, graph_)].merge(
                graph_[boost::target(*out_edge_iter_func_seg, graph_)]);

            // we are iterating over the predecessors of the C
            Graph::out_edge_iterator out_edge_begin_scc_c,
              out_edge_end_scc_c, out_edge_iter_scc_c;

            boost::tie(out_edge_begin_scc_c, out_edge_end_scc_c) =
              boost::out_edges(boost::target(*out_edge_iter_func_seg, graph_),
                         graph_);

            for(out_edge_iter_scc_c  = out_edge_begin_scc_c;
                out_edge_iter_scc_c != out_edge_end_scc_c;
                out_edge_iter_scc_c++) {

              // Adding an edge from  A to the successor of C
              if(!boost::edge(boost::source(*in_edge_iter,graph_),
                              boost::target(*out_edge_iter_scc_c, graph_),
                              graph_).second) {

                boost::add_edge(
                           boost::source(*in_edge_iter,graph_),
                           boost::target(*out_edge_iter_scc_c, graph_),
                           graph_);
              }
            }
          }
        }
      }

      // In the following we are handing the returning of the recursive called
      // functions
      std::set<VertexDescriptor> to_delete;

      boost::tie(out_edge_begin, out_edge_end) =
        boost::out_edges(*recursive_iter, graph_);
      for(out_edge_iter  = out_edge_begin;
          out_edge_iter != out_edge_end;
          out_edge_iter++) {

        VertexDescriptor successor = boost::target(*out_edge_iter, graph_);

        Graph::out_edge_iterator follow_out_edge_begin, follow_out_edge_end,
          follow_out_edge_iter;
        boost::tie(follow_out_edge_begin, follow_out_edge_end) =
          boost::out_edges(successor, graph_);

        for(follow_out_edge_iter  = follow_out_edge_begin;
            follow_out_edge_iter != follow_out_edge_end;
            follow_out_edge_iter++) {

          for(SegmentSet::iterator
              leaving_seg_iter  = precached_function->leaving_segments_.begin();
              leaving_seg_iter != precached_function->leaving_segments_.end();
              leaving_seg_iter++) {

            graph_[*leaving_seg_iter].merge(graph_[successor]);

            if(!boost::edge(*leaving_seg_iter,
                           boost::target(*follow_out_edge_iter, graph_),
                           graph_).second) {

              boost::add_edge(*leaving_seg_iter,
                         boost::target(*follow_out_edge_iter, graph_),
                         graph_);
            }
          }
        }

        to_delete.insert(boost::target(*out_edge_iter, graph_));
      }

      // here, we are deleting all the temporally segments
      for(SegmentSet::iterator iter = to_delete.begin();
          iter != to_delete.end(); iter++) {

        boost::clear_vertex(*iter, graph_);
        boost::remove_vertex(*iter, graph_);
      }

      boost::clear_vertex(*recursive_iter, graph_);
      boost::remove_vertex(*recursive_iter, graph_);
    }
  }
  recursive_function_calls_.clear();
}

void
risc::sg::SegmentGraph::
separate_variable_declaration_and_initializer_on_demand(SgBasicBlock *&bb,
    bool duplicate_segments)
{
  // We are iterating over all
  std::set<SgVariableDeclaration*> need_split;
  for(std::vector<SgStatement*>::iterator
      stmt_iter  = bb->get_statements().begin();
      stmt_iter != bb->get_statements().end();
      stmt_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*stmt_iter);

    // In this case we are facing a variable declaration. We have to be
    // careful
    if(var_decl) {

      SgInitializedName *initialized_name = var_decl->get_decl_item(0);

      // In this case the variable declaration is associated with a
      // initialization. So we have check if any function call occurs in the
      // initialization process
      if(initialized_name->get_initptr()) {

        // We search for all function calls.
        std::vector<SgNode*> func_calls;

        if(isSgAssignInitializer(initialized_name->get_initptr())) {
          func_calls =
            NodeQuery::querySubTree(isSgAssignInitializer(
                           initialized_name->get_initptr())->get_operand(),
                               V_SgFunctionCallExp);
        }

        if(isSgAggregateInitializer(initialized_name->get_initptr())) {

          func_calls =
            NodeQuery::querySubTree(isSgAggregateInitializer(
                           initialized_name->get_initptr())->get_initializers(),
                               V_SgFunctionCallExp);
        }

        // Now, we can iterate over all function calls and check if one of
        // the called functions has some boundary statements inside
        for(std::vector<SgNode*>::iterator
            func_call_iter  = func_calls.begin();
            func_call_iter != func_calls.end();
            func_call_iter++) {

          SgFunctionCallExp *func_call = isSgFunctionCallExp(*func_call_iter);

          // Here we are checking, if the function has some annotation.
          // If so, we can return and we don't have to analyze the function
          FunctionAnnotationAttributes *annotation
            = get_function_call_annotation(func_call);

          if(annotation) {
            if(annotation->has_boundary()) {
              need_split.insert(var_decl);
            }
            continue;
          }

          // Okay, the function has a definition and no pragma statement.
          // So, we have to analyze the function.
          // We are getting the function definition of the associated function
          // call. So, we can check if this function was already pre cached.
          SgFunctionDeclaration *func_decl
            = tools::get_first_non_defining_function_declaration(func_call);

          if(func_decl) {

            if(follow_function_call(func_decl)) {

              func_decl = isSgFunctionDeclaration(
                              func_decl->get_definingDeclaration());

              SgFunctionDefinition *func_def = func_decl->get_definition();

              // we are loading the pre cached information
              CachedFunctionAstAttributes *precached_function =
                dynamic_cast<CachedFunctionAstAttributes*>(
                    func_def->getAttribute("precached_info"));

              // If the precached_function pointer is NULL, the function was not
              // analyzed yet. So, we have to do it right now
              if(!precached_function) {

                // Here, we are adding the arguments which are passed by
                // reference to the corresponding parameter in the function
                // declaration
                // We need this information later for the conflict analysis
                add_mapped_symbol_to_reference(func_call, func_def);

                precached_function = build_segment_graph_for_function(func_def,
                                         duplicate_segments);
              } else {

                // This function is already analyzed and we don't want to
                // analyze it twice (decomposition etc.). So, we are following
                // the function call and we are adding arguments passed by
                // reference to the definition. Afterwards, we get all the
                // function call expressions and forward references recursively.
                forward_reference_function_parameters(func_call);
              }

              // Here we check if a boundary statement is included in the
              // function call
              if(precached_function->has_boundary()) {
                need_split.insert(var_decl);
              }

              // Also we check if the function call is a direct boundary statement
              for(FunctionCallBoundarySet::const_iterator
                  func_def_iter  = function_boundaries_.begin();
                  func_def_iter != function_boundaries_.end();
                  func_def_iter++) {

                if(func_call->getAssociatedFunctionSymbol() ==
                    (*func_def_iter)->get_declaration()->
                    search_for_symbol_from_symbol_table()) {
                  need_split.insert(var_decl);
                }
              }
            }
          }
        }
      }
    }
  }

  // In the following, we are inserting after each identified variable
  // declaration an assignment operator. The assignment operator overtakes
  // the assignment initializer expression and assigns it to the variable.
  for(std::set<SgVariableDeclaration*>::iterator
      var_decl_iter  = need_split.begin();
      var_decl_iter != need_split.end();
      var_decl_iter++) {

    // we need this variable reference expression for the assignment operator
    SgVarRefExp *var_ref = SageBuilder::buildVarRefExp(*var_decl_iter);

    // getting the assignment initializer expression
    SgInitializedName *initialized_name = (*var_decl_iter)->get_decl_item(0);
    SgAssignInitializer *assign_init =
      isSgAssignInitializer(initialized_name->get_initptr());

    if(assign_init) {

      SgExpression *assignment =
        isSgExpression(SageInterface::deepCopyNode(assign_init->get_operand()));

      // Here we build the assignment expression with associated statement
      SgExprStatement *expr_stmt =
        SageBuilder::buildAssignStatement(var_ref,
        assignment);

      // Inserting the expression statement after the variable declaration
      SageInterface::insertStatement(*var_decl_iter, expr_stmt, false);

      // we are deleting the assignment initialization of the variable
      // declaration
      initialized_name->set_initptr(NULL);
    }

    // In this case we have something like
    // int array[] = {1, foo(), 3}; where 'foo()' contains a boundary
    // statement. We have to transform like
    // int temp_foo();
    // int array[] = {1, temp_foo, 3};
    SgAggregateInitializer *aggregate_initializer =
      isSgAggregateInitializer(initialized_name->get_initptr());
    if(aggregate_initializer) {

      SgExprListExp *expr_list = aggregate_initializer->get_initializers();

      SgExpressionPtrList &initializer_expressions =
        expr_list->get_expressions();
      for(std::vector<SgExpression*>::iterator
          expr_iter  = initializer_expressions.begin();
          expr_iter != initializer_expressions.end();
          expr_iter++) {

        std::vector<SgNode*> func_calls =
          NodeQuery::querySubTree(*expr_iter, V_SgFunctionCallExp);

        for(std::vector<SgNode*>::iterator
            func_call_iter  = func_calls.begin();
            func_call_iter != func_calls.end();
            func_call_iter++) {

          SgFunctionCallExp *func_call = isSgFunctionCallExp(*func_call_iter);

          FunctionAnnotationAttributes *annotation
            = get_function_call_annotation(func_call);

          SgFunctionDeclaration *func_decl
            = tools::get_first_non_defining_function_declaration(func_call);

          if(func_decl) {

            bool has_boundary = false;
            if(annotation) {
              if(annotation->has_boundary()) {
                has_boundary = true;
              }
            } else {

              func_decl = isSgFunctionDeclaration(
                              func_decl->get_definingDeclaration());

              SgFunctionDefinition *func_def = func_decl->get_definition();

              CachedFunctionAstAttributes *precached_function =
                dynamic_cast<CachedFunctionAstAttributes*>(
                    func_def->getAttribute("precached_info"));

              has_boundary = precached_function->has_boundary();
            }

            if(has_boundary) {

              // In this situation we are facing a situation like this:
              // 'int array[] = {1, foo(), 3};' where 'foo()' contains a
              // boundary statement. We have to transform this statement into:
              // int temp_foo = foo();
              // int array[] = {1, temp_foo, 3};
              // In particular, we are iterating currently of foo() in the
              // expression list

              // copy of the call expression
              SgExpression *assignment =
                isSgExpression(SageInterface::deepCopyNode(func_call));

              // here we create 'int temp_foo = foo()'
              SgVariableDeclaration *var_decl =
                SageBuilder::buildVariableDeclaration(
                                 generate_unique_name(func_call),
                                 func_decl->get_orig_return_type(),
                                 NULL,
                                 (*var_decl_iter)->get_scope());

              SageInterface::insertStatement(*var_decl_iter, var_decl);
              SgVarRefExp *var_ref_exp = SageBuilder::buildVarRefExp(var_decl);

              SgExprStatement *expr_stmt =
                SageBuilder::buildAssignStatement(var_ref_exp, assignment);

              // Insert after the new variable declaration
              SageInterface::insertStatement(var_decl, expr_stmt, false);
              SageInterface::replaceExpression(func_call, var_ref_exp);
            }
          }
        }
      }
    }
  }
}

void
risc::sg::SegmentGraph::
create_temp_variable_for_expression(SgExpression *expr, SgBasicBlock *&bb)
{
  SgStatement *stmt = SageInterface::getEnclosingStatement(expr);
  SgScopeStatement *scope = stmt->get_scope();

  SgExpression *assignment = isSgExpression(SageInterface::deepCopyNode(expr));

  SgVariableDeclaration *result =
    SageBuilder::buildVariableDeclaration(generate_unique_name(expr),
                     expr->get_type(),
                     NULL,
                     scope);

  SgVarRefExp *var_ref = SageBuilder::buildVarRefExp(result);

  SgExprStatement *expr_stmt = SageBuilder::buildAssignStatement(
      SageBuilder::buildVarRefExp(result),
      assignment);

  SageInterface::prependStatement(result, bb);

  // We are adding the expression on the second last position
  SgStatement *last_in_bb = (*bb).get_statements().back();
  SageInterface::insertStatement(last_in_bb, expr_stmt, true);
  SageInterface::replaceExpression(expr, var_ref);
}

void
risc::sg::SegmentGraph::
decompose_expression_with_boundary_calls(SgExpression *expr, SgBasicBlock *&bb,
    bool duplicate_segments)
{
  if(!should_decompose_function(expr, duplicate_segments)) {
    return;
  }

  SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expr);
  if(func_call_exp) {

    SgExprListExp *arguments = func_call_exp->get_args();
    for(SgExpressionPtrList::iterator
        expr_iter  = arguments->get_expressions().begin();
        expr_iter != arguments->get_expressions().end();
        expr_iter++) {

      decompose_expression_with_boundary_calls(*expr_iter,bb,
          duplicate_segments);
    }

    // Here we are checking, if the function has some annotation.
    // If so, we can return and we don't have to analyze the function

    FunctionAnnotationAttributes *annotation
      = get_function_call_annotation(func_call_exp);
    if(annotation) {

      SgExprStatement *expr_stmt =
        isSgExprStatement(SageInterface::getEnclosingStatement(func_call_exp));

      if(annotation->has_boundary() &&
          expr_stmt->get_expression() != func_call_exp) {

        if(bb == NULL) {

          SgStatement *stmt = SageInterface::getEnclosingStatement(expr);
          bb = SageBuilder::buildBasicBlock();
          SageInterface::replaceStatement(stmt, bb);
          SageInterface::appendStatement(stmt, bb);
        }
        create_temp_variable_for_expression(expr, bb);
      }

      return;
    }

    // Okay, the function has a definition and no pragma statement.
    // So, we have to analyze the function
    SgFunctionDeclaration *func_decl
      = tools::get_first_non_defining_function_declaration(func_call_exp);

    if(func_decl) {
      func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());

      SgFunctionDefinition *func_def = NULL;

      // Here, we check if the function call is a direct boundary statement
      // If yes, then we can return
      SgStatement *enclosing_statement =
        SageInterface::getEnclosingStatement(func_call_exp);

      for(FunctionCallBoundarySet::const_iterator
          func_def_iter  = function_boundaries_.begin();
          func_def_iter != function_boundaries_.end();
          func_def_iter++) {

        if(func_call_exp->getAssociatedFunctionSymbol() ==
            (*func_def_iter)->get_declaration()->
              search_for_symbol_from_symbol_table()) {
          return;
        }
      }

      // In this case we have the body of the function
      if(func_decl) {

        if(follow_function_call(func_decl)) {

          func_def = func_decl->get_definition();

          // we are loading the pre cached information
          CachedFunctionAstAttributes *precached_function =
            dynamic_cast<CachedFunctionAstAttributes*>(
                func_def->getAttribute("precached_info"));

          // If the precached_function pointer is NULL, the function was not
          // analyzed yet. So, we have to do it right now
          if(!precached_function) {

            // Here, we are adding the arguments which are passed by reference
            // to the corresponding parameter in the function declaration
            // We need this information later for the conflict analysis
            add_mapped_symbol_to_reference(func_call_exp, func_def);

            precached_function = build_segment_graph_for_function(func_def,
                                     duplicate_segments);

          } else {

            // This function is already analyzed and we don't want to analyze
            // it twice (decomposition etc.). So, we are following the
            // function call and we are adding arguments passed by reference
            // to the definition. Afterwards, we get all the function call
            // expressions and forward references recursively.
            forward_reference_function_parameters(func_call_exp);
          }

          SgExprStatement *expr_stmt =
            isSgExprStatement(
                SageInterface::getEnclosingStatement(func_call_exp));

          if(precached_function->has_boundary() &&
              expr_stmt->get_expression() != func_call_exp) {

            if(bb == NULL) {

              SgStatement *stmt = SageInterface::getEnclosingStatement(expr);
              bb = SageBuilder::buildBasicBlock();
              SageInterface::replaceStatement(stmt, bb);
              SageInterface::appendStatement(stmt, bb);
            }
            create_temp_variable_for_expression(expr, bb);
          }
        }
      }
    }
  }

  #define DECOMPOSE_EXPR_IN_BINARY_OP(Operator)                          \
  if(is##Operator(expr)) {                                               \
    Operator *op = is##Operator(expr);                                   \
    decompose_expression_with_boundary_calls(op->get_lhs_operand(), bb,  \
        duplicate_segments);                                             \
    decompose_expression_with_boundary_calls(op->get_rhs_operand(), bb,  \
        duplicate_segments);                                             \
    return;                                                              \
  }
  DECOMPOSE_EXPR_IN_BINARY_OP(SgAndAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgIorAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgXorAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgLshiftAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgRshiftAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgPlusAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgMinusAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgMultAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgDivAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgModAssignOp);

  DECOMPOSE_EXPR_IN_BINARY_OP(SgLessThanOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgGreaterThanOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgLessOrEqualOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgGreaterOrEqualOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgEqualityOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgNotEqualOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgLshiftOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgRshiftOp);

  DECOMPOSE_EXPR_IN_BINARY_OP(SgBitAndOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgBitOrOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgAndOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgOrOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgBitXorOp);

  DECOMPOSE_EXPR_IN_BINARY_OP(SgAddOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgSubtractOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgDivideOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgMultiplyOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgModOp);

  DECOMPOSE_EXPR_IN_BINARY_OP(SgCommaOpExp);

  DECOMPOSE_EXPR_IN_BINARY_OP(SgAssignOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgPntrArrRefExp);

  DECOMPOSE_EXPR_IN_BINARY_OP(SgDotStarOp);
  DECOMPOSE_EXPR_IN_BINARY_OP(SgArrowStarOp);
  #undef DECOMPOSE_EXPR_IN_BINARY_OP

  #define DECOMPOSE_EXPR_IN_UNARY_OP(Operator)                       \
  if(is##Operator(expr)) {                                           \
    Operator *op = is##Operator(expr);                               \
    decompose_expression_with_boundary_calls(op->get_operand(), bb,  \
        duplicate_segments);                                         \
    return;                                                          \
  }
  DECOMPOSE_EXPR_IN_UNARY_OP(SgUnaryAddOp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgMinusOp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgNotOp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgBitComplementOp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgCastExp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgPointerDerefExp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgAddressOfOp);

  DECOMPOSE_EXPR_IN_UNARY_OP(SgPlusPlusOp);
  DECOMPOSE_EXPR_IN_UNARY_OP(SgMinusMinusOp);
  #undef DECOMPOSE_EXPR_IN_UNARY_OP

  //We have to document this case off otherwise it would be result in an
  //infinity loop. Let's say we have 'int x = foo()' and foo() has a boundary
  //statement. So, we would replace it recursively and recursively and ...
  //This case is already handled in the function
  //'separate_variable_declaration_and_initializer_on_demand'
  //
  //DECOMPOSE_EXPR_IN_UNARY_OP(SgAssignInitializer);

  if(isSgVarRefExp(expr)) {
  }

  if(isSgArrowExp(expr)) {
    SgArrowExp *arrow_exp = isSgArrowExp(expr);
    decompose_expression_with_boundary_calls(
        arrow_exp->get_rhs_operand(),
        bb,
        duplicate_segments);
  }

  if(isSgDotExp(expr)) {
    SgDotExp *arrow_exp = isSgDotExp(expr);
    decompose_expression_with_boundary_calls(
        arrow_exp->get_rhs_operand(),
        bb,
        duplicate_segments);
  }

  if(isSgSizeOfOp(expr)) {
    SgSizeOfOp *size_of_op = isSgSizeOfOp(expr);

    // Not necessarily we have an expression pointer. For instance,
    // 'sizeof(int)' does not have one.
    if(size_of_op->get_operand_expr()) {

      decompose_expression_with_boundary_calls(size_of_op->get_operand_expr(),
        bb, duplicate_segments);
    }
  }

  if(isSgConditionalExp(expr)) {
    SgConditionalExp *conditional_exp = isSgConditionalExp(expr);

    decompose_expression_with_boundary_calls(
        conditional_exp->get_conditional_exp(), bb, duplicate_segments);
    decompose_expression_with_boundary_calls(
        conditional_exp->get_true_exp(), bb, duplicate_segments);
    decompose_expression_with_boundary_calls(
        conditional_exp->get_false_exp(),bb, duplicate_segments);
  }

  assert(true && "Operator not supported");
}

std::string
risc::sg::SegmentGraph::generate_unique_name(SgNode *node)
{
  SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(node);
  std::ostringstream name_stream;

  SgFunctionDeclaration *func_decl
    = tools::get_first_non_defining_function_declaration(func_call_exp);

  name_stream << "__risc_temp_" << counter << "_"
              << SageInterface::get_name(func_decl);

  std::string name(name_stream.str());
  counter++;

  return name;
}

risc::sg::CachedFunctionAstAttributes*
risc::sg::SegmentGraph::build_segment_graph_for_function(
    SgFunctionDefinition *func_def, bool duplicate_segments)
{
  Segment function_segment(NULL, false);
  VertexDescriptor function_vertex
    = boost::add_vertex(function_segment, graph_);

  precached_function_segments_.insert(function_vertex);

  CachedFunctionAstAttributes *precached_function
    = new CachedFunctionAstAttributes(function_vertex,
            CachedFunctionAstAttributes::Connecting);
  func_def->setAttribute("precached_info", precached_function);

  tools::AstAttributeColor *call_hierarchy_color
    = new tools::AstAttributeColor();

  func_def->setAttribute("function_call_hierarchy", call_hierarchy_color);
  call_hierarchy_color->call_hierarchy_ =
    tools::AstAttributeColor::IN_CALL_HIERARCHY;

  SegmentSet current_segments;
  SegmentSet break_segments;
  SegmentSet continue_segments;
  current_segments.insert(function_vertex);

  SegmentSet func_segments = build_graph(func_def->get_body(),
                                 current_segments,
                                 break_segments,
                                 continue_segments,
                                 duplicate_segments);

  precached_function->leaving_segments_.insert(
                                            func_segments.begin(),
                                            func_segments.end());

  call_hierarchy_color->call_hierarchy_ =
    tools::AstAttributeColor::NOT_IN_CALL_HIERARCHY;

  // If we analyze a channel function, the related segments  will be later
  // cloned. So, we have to put them into the set 'precached_function_segments_'
  // for later cleanup
  SgMemberFunctionDeclaration *mem_func_decl
    = isSgMemberFunctionDeclaration(func_def->get_declaration());
  if(mem_func_decl) {
    if(risc::get_design()->find_prim_channel_definition(
          mem_func_decl->get_class_scope())
        || risc::get_design()->find_channel_definition(
              mem_func_decl->get_class_scope())) {

      std::list<int> reachable_segs
        = risc::sg::get_all_reachable_segments(*this, function_segment.id_);

      for(std::list<int>::iterator
           iter  = reachable_segs.begin();
           iter != reachable_segs.end();
           iter++) {
        precached_function_segments_.insert(id_to_vertex_descriptor(*iter));
      }
    }
  }

  return precached_function;
}

void
risc::sg::SegmentGraph::set_all_segments_to_untouched()
{

  std::pair<VertexIterator, VertexIterator> vp;

  VertexIterator starting_vertex;

  for(vp = boost::vertices(graph_); vp.first != vp.second; ++vp.first) {
    graph_[*(vp.first)].helper_for_graph_algorithms_
      = Segment::Untouched;
  }
}

risc::sg::FunctionAnnotationAttributes*
risc::sg::SegmentGraph::get_function_call_annotation(
    SgFunctionCallExp *func_call_exp)
{
  // we don't want to dig into boundary statements
  for(FunctionCallBoundarySet::const_iterator
      func_def_iter  = function_boundaries_.begin();
      func_def_iter != function_boundaries_.end();
      func_def_iter++) {

    if(func_call_exp->getAssociatedFunctionSymbol() ==
        (*func_def_iter)->get_declaration()->
          search_for_symbol_from_symbol_table()) {
      return NULL;
    }
  }

  SgFunctionDeclaration *func_decl
    = tools::get_first_non_defining_function_declaration(func_call_exp);

  FunctionAnnotationAttributes *function_annotation_attribute = NULL;
  //FunctionAnnotationAttributes *function_annotation_attribute =
  //  dynamic_cast<FunctionAnnotationAttributes*>(
  //      func_decl->getAttribute("function_annotation"));

  // In the following if statement, we are getting the pragma of the
  // interface function declaration
  bool is_port_call = tools::function_call_belongs_to_a_port(func_call_exp);
  FunctionAnnotation *interface_attribute = NULL;
  FunctionAnnotation *function_decl_attribute = NULL;
  SgFunctionDeclaration *interface_func_decl = NULL;

  // first try to the function annotation
  if(is_port_call) {

    // first, we get the annotation of the interface
    SgBinaryOp *binary = isSgBinaryOp(func_call_exp->get_function());

    if(binary) {

      SgMemberFunctionSymbol *func_symbol
        = isSgMemberFunctionRefExp(binary->get_rhs_operand())->get_symbol();
      interface_func_decl = func_symbol->get_declaration();
      interface_attribute
        = analyze_prepended_pragmas_for_annotation(interface_func_decl);
    }

    // next, we get the annotation of the channel function declaration
    function_decl_attribute
      = analyze_prepended_pragmas_for_annotation(func_decl);

    if(interface_attribute == NULL && function_decl_attribute != NULL) {
      std::cout << "Error: The interface in line "
                << interface_func_decl->get_file_info()->get_line()
                << " has no annotation,"
                << " however, the associated implementation in line "
                << func_decl->get_file_info()->get_line()
                << " has one." << std::endl
                << "The implementation annotation must match the interface"
                << " annotation."
                << std::endl;
      exit(1);
    }

  } else {

    // here we get the annotation of the regular function

    // func_decl can be NULL if it is a virtual function call
    if(func_decl == NULL) {
      return NULL;
    }

    function_decl_attribute
      = analyze_prepended_pragmas_for_annotation(func_decl);
  }


  // This is the case that we have no comment, however, the function
  // has a definition. So we skip this one
  if(!function_annotation && func_decl->get_definingDeclaration()) {
    return NULL;
  }


  // Finally, we attach the annotation to the declaration
  if(is_port_call) {
    if(interface_attribute != NULL) {

      if(interface_attribute) {

        function_annotation_attribute
          = new FunctionAnnotationAttributes(interface_attribute);
        interface_func_decl->setAttribute("function_annotation",
                       function_annotation_attribute);


        function_annotation_attribute
          = new FunctionAnnotationAttributes(interface_attribute);
        func_decl->setAttribute("function_annotation",
                       function_annotation_attribute);

      } else {

        // here we catch the case of a undocumented function
        char name[strlen(SageInterface::get_name(func_decl).c_str())];
        strcpy(name, SageInterface::get_name(func_decl).c_str());

        AnnotatedTime *annotated_time = new AnnotatedTime();

        function_annotation_attribute
          = new FunctionAnnotationAttributes(name, true, NO_WAIT,
                    annotated_time);

        interface_func_decl->setAttribute("function_annotation",
                       function_annotation_attribute);

        name[strlen(SageInterface::get_name(func_decl).c_str())];
        strcpy(name, SageInterface::get_name(func_decl).c_str());

        annotated_time = new AnnotatedTime();

        function_annotation_attribute
          = new FunctionAnnotationAttributes(name, true, NO_WAIT,
                    annotated_time);

        func_decl->setAttribute("function_annotation",
                       function_annotation_attribute);
      }
    }

  } else {
    if(function_annotation_attribute == NULL) {

      if(function_decl_attribute) {

        function_annotation_attribute
          = new FunctionAnnotationAttributes(function_decl_attribute);
        func_decl->setAttribute("function_annotation",
                       function_annotation_attribute);

      } else {

        // here we catch the case of a undocumented function
        char name[strlen(SageInterface::get_name(func_decl).c_str())];
        strcpy(name, SageInterface::get_name(func_decl).c_str());

        AnnotatedTime *annotated_time = new AnnotatedTime();

        function_annotation_attribute
          = new FunctionAnnotationAttributes(name, true, NO_WAIT,
                    annotated_time);

        func_decl->setAttribute("function_annotation",
                       function_annotation_attribute);
      }
    }
  }

  return function_annotation_attribute;
}

risc::sg::CachedFunctionAstAttributes*
risc::sg::SegmentGraph::create_cached_function_attribute_for_annotated_function(
    SgNode* boundary_node,
    WAIT_CONSTRUCT wait_type,
    SegmentSet incoming_segments,
    bool is_conflict_free)
{
  // we can take here the entire expression because everything is already
  // decomposed for the surrounding expression
  add_expression_to_segment(incoming_segments,
      SageInterface::getEnclosingStatement(boundary_node));

  if(wait_type == NO_WAIT) {
    return NULL;
  }

  Segment segment(boundary_node, true);
  VertexDescriptor segment_vertex = boost::add_vertex(segment, graph_);

  graph_[segment_vertex].calls_conflict_function_ = !is_conflict_free;

  // TODO We need a real debugging flag!
  //if(debugging_) {
    Conflict conflict(boundary_node);
    graph_[segment_vertex].conflicts_.insert(conflict);
  //}

  CachedFunctionAstAttributes*  precached_function =
    new CachedFunctionAstAttributes(segment_vertex,
            CachedFunctionAstAttributes::Inlining);
  boundary_node->setAttribute("precached_info", precached_function);

  // linking the new segment
  for(SegmentSet::iterator
      iter  = incoming_segments.begin();
      iter != incoming_segments.end();
      iter++) {

    boost::add_edge(*iter, segment_vertex, graph_);
  }

  SegmentGraph::SegmentSet temp;
  temp.insert(segment_vertex);

  switch(wait_type) {

    case UNCONDITIONAL_WAIT:
      break;

    case CONDITIONAL_WAIT:
      temp.insert(incoming_segments.begin(), incoming_segments.end());
      break;

    case LOOPED_WAIT:
      temp.insert(incoming_segments.begin(), incoming_segments.end());
      boost::add_edge(segment_vertex, segment_vertex, graph_);
      break;

    case WORST_WAIT:
      temp.insert(incoming_segments.begin(), incoming_segments.end());
      boost::add_edge(segment_vertex, segment_vertex, graph_);
      break;
  }

  precached_function->leaving_segments_ = temp;

  return precached_function;
}

void
risc::sg::SegmentGraph::add_mapped_symbol_to_reference(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def)
{
  SgFunctionDeclaration *func_decl
    = isSgFunctionDeclaration(
          func_def->get_declaration()->get_definingDeclaration());

  // we are iterating over all arguments in the function call
  int counter = 0;
  for(SgExpressionPtrList::iterator
      arg_iter  = func_call->get_args()->get_expressions().begin();
      arg_iter != func_call->get_args()->get_expressions().end();
      arg_iter++, counter++) {

    SgVariableSymbol *parameter
      = isSgVariableSymbol(
            func_decl->get_args().at(counter)
              ->search_for_symbol_from_symbol_table());

    // if it is a reference, we are adding the argument to the symbol
    // of the parameter
    if(SageInterface::isReferenceType(parameter->get_type())) {

      FunctionParameterReferenceAttribute *reference_annotation =
        dynamic_cast<FunctionParameterReferenceAttribute*>(
            parameter->getAttribute("parameter_reference"));

      if(reference_annotation == NULL) {
        reference_annotation = new FunctionParameterReferenceAttribute();
        parameter->setAttribute("parameter_reference", reference_annotation);
      }

      // We are calling get_most_outer_var to over something like
      // foo(array[i]). Where we are adding the symbol of array
      SgVariableSymbol *var = tools::get_most_outer_var(*arg_iter);

      // we have to check if there is a symbol at all. For instance a function
      // call which passes only a literal has no symbol
      if(var) {
        if(SageInterface::isReferenceType(var->get_type())) {

          // In this case, the function argument is a reference. So, we have to
          // find the mapped variable

          if(tools::is_function_parameter_variable(var)) {

            FunctionParameterReferenceAttribute *argument_reference_attributes
              = dynamic_cast<FunctionParameterReferenceAttribute*>(
                    var->getAttribute("parameter_reference"));

            reference_annotation->mapped_variables_.insert(
                argument_reference_attributes->mapped_variables_.begin(),
                argument_reference_attributes->mapped_variables_.end());

          } else {

            LocalReferenceAttribute *local_mapping
              = dynamic_cast<LocalReferenceAttribute*>(
                    var->getAttribute("local_reference"));

            SgVariableSymbol *temp = local_mapping->mapped_variable_;

            // The function argument is function parameter as well.
            if(tools::is_function_parameter_variable(temp)) {
              FunctionParameterReferenceAttribute *argument_reference_attributes
                = dynamic_cast<FunctionParameterReferenceAttribute*>(
                      temp->getAttribute("parameter_reference"));

              reference_annotation->mapped_variables_.insert(
                  argument_reference_attributes->mapped_variables_.begin(),
                  argument_reference_attributes->mapped_variables_.end());

            } else {
              reference_annotation->mapped_variables_.insert(temp);
            }
          }

        } else {

          SgVariableSymbol *temp = tools::get_most_outer_var(*arg_iter);

          // we have to check if there is a symbol at all. For instance a
          // function call which passes only a literal has no symbol
          if(temp) {
            reference_annotation->mapped_variables_.insert(temp);
          }
        }
      }
    }
  }
}

void
risc::sg::SegmentGraph::forward_reference_function_parameters(
    SgFunctionCallExp *func_call_exp)
{
  // We only follow functions where we have the definition and which are not
  // annotated.
  // Also, we don't follow any function calls into the std namespace or
  // in the systemc library
  FunctionAnnotationAttributes *annotation =
    get_function_call_annotation(func_call_exp);

  if(annotation) {
    return;
  }

  // This means we have a function call and we can determine the called
  // function definition
  SgFunctionDeclaration *func_decl
    = tools::get_first_non_defining_function_declaration(func_call_exp);

  if(func_decl) {
    // we don't have to follow functions which don't take any argument by
    // reference
    if(tools::has_reference_parameter(func_decl) == false) {
      return;
    }

    if(tools::is_in_std_namespace(func_decl)
        || tools::is_in_systemc_namespace(func_decl)
        || tools::is_in_gnu_cxx_namespace(func_decl)) {
      return;
    }

    func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());
    SgFunctionDefinition *func_def = func_decl->get_definition();

    // Here, we are adding the references to the function definition
    add_mapped_symbol_to_reference(func_call_exp, func_def);

    // Here, we are traversing recursively the function definition. If we find
    // another function call, we will follow this function call
    std::vector<SgNode*> func_calls
      = NodeQuery::querySubTree(func_def, V_SgFunctionCallExp);

    for(std::vector<SgNode*>::iterator
        iter  = func_calls.begin();
        iter != func_calls.end();
        iter++) {

      SgFunctionCallExp *func_call = isSgFunctionCallExp(*iter);
      forward_reference_function_parameters(func_call);
    }
  }
}

FunctionAnnotation*
risc::sg::SegmentGraph::analyze_prepended_pragmas_for_annotation(
    SgFunctionDeclaration *func_decl)
{
#if 0
  // This is commented off because it would slow down the compiler
  // Also, I am not sure how stable this one is
  std::vector<SgFunctionDeclaration*> decls
    = tools::get_all_other_function_declarations(func_decl);
  for(std::vector<SgFunctionDeclaration*>::iterator
      iter  = decls.begin();
      iter != decls.end();
      iter++) {
    bool found = false;
#endif

    // we have to reset this to null because
    function_annotation = NULL;

    SgClassDefinition *scope
      = isSgClassDefinition(SageInterface::getEnclosingScope(func_decl));
    SgStatement *previous = func_decl;

    SgDeclarationStatementPtrList *statements = NULL;

    // here we get the surrounding statement list of the function declaration
    if(isSgClassDefinition(SageInterface::getEnclosingScope(func_decl))) {
      SgDeclarationStatementPtrList &temp
        = isSgClassDefinition(
              SageInterface::getEnclosingScope(func_decl))->get_members();
      statements = &temp;
    } else {

      if(isSgGlobal(SageInterface::getEnclosingScope(func_decl))) {
        SgDeclarationStatementPtrList &temp
          = isSgGlobal(
                SageInterface::getEnclosingScope(func_decl))
              ->get_declarations();
        statements = &temp;
      } else {

        if(isSgNamespaceDefinitionStatement(
              SageInterface::getEnclosingScope(func_decl))) {
          SgDeclarationStatementPtrList &temp
            = isSgNamespaceDefinitionStatement(
                  SageInterface::getEnclosingScope(func_decl))
                ->get_declarations();
          statements = &temp;
        }
      }
    }

    // get the previous statement
    for(SgDeclarationStatementPtrList::iterator
        iter  = statements->begin();
        iter != statements->end();
        iter++) {

      if(func_decl == (*iter)->get_firstNondefiningDeclaration()) {

        if(statements->begin() != iter) {

          previous = *(iter-1);
          break;
        }
      }
    }

    SgPragmaDeclaration *pragma = isSgPragmaDeclaration(previous);

    while(pragma) {

      function_annotation_parser_scan_string(
          pragma->get_pragma()->get_pragma().c_str());
      function_annotation_parserparse();

      if(function_annotation) {
        // belongs to previously commented off function
        //func_decl = *iter;
        return function_annotation;
        //break;
      }

      // get the previous statement
      for(SgDeclarationStatementPtrList::iterator
          iter  = statements->begin();
          iter != statements->end();
          iter++) {

        if(previous == *iter) {

          previous = *(iter-1);
          break;
        }
      }
      pragma = isSgPragmaDeclaration(previous);
    }
#if 0
    //belongs to previously commented off function
    if(found) {
      break;
    }
  }
#endif
}

void
risc::sg::SegmentGraph::mark_not_conflict_free_func_calls_in_segment(
    SgNode *stmt, SegmentSet segments)
{
  if(stmt == NULL) {
    return;
  }

  std::vector<SgNode*> func_calls =
    NodeQuery::querySubTree(stmt , V_SgFunctionCallExp);

  for(std::vector<SgNode*>::iterator
      func_call_iter  = func_calls.begin();
      func_call_iter != func_calls.end();
      func_call_iter++) {

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(*func_call_iter);

    FunctionAnnotationAttributes *function_annotation
       = get_function_call_annotation(func_call_exp);

    if(function_annotation) {
      if(function_annotation->is_conflict_free_ == false) {

        for(SegmentSet::iterator
            seg_iter  = segments.begin();
            seg_iter != segments.end();
            seg_iter++) {

          graph_[*seg_iter].calls_conflict_function_ = true;

          // TODO We need a real debugging flag!
          //if(debugging_) {
            Conflict conflict(func_call_exp);
            graph_[*seg_iter].conflicts_.insert(conflict);
          //}
        }
      }
    }
  }
}

bool
risc::sg::SegmentGraph::follow_function_call(SgFunctionDeclaration *func_decl)
{
  return tools::is_in_std_namespace(func_decl) == false
    && tools::is_in_gnu_cxx_namespace(func_decl) == false
    && tools::is_in_systemc_namespace(func_decl) == false;
}

bool
risc::sg::SegmentGraph::has_transition(int seg_id_from, int seg_id_to)
{
  Graph::edge_iterator edge_iter_begin, edge_iter_end;
  boost::tie(edge_iter_begin, edge_iter_end) = boost::edges(graph_);

  for(; edge_iter_begin != edge_iter_end; edge_iter_begin++) {
    if(graph_[boost::source(*edge_iter_begin, graph_)].id_ == seg_id_from
        && graph_[boost::target(*edge_iter_begin, graph_)].id_ == seg_id_to) {
      return true;
    }
  }

  return false;
}

risc::sg::VertexDescriptor
risc::sg::SegmentGraph::id_to_vertex_descriptor(int seg_id)
{
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(graph_);
      vp.first != vp.second; ++vp.first) {
    if(graph_[*vp.first].id_ == seg_id) {
      return *vp.first;
    }
  }

  assert(NULL);
  VertexDescriptor result;
  return result;
}

int risc::sg::SegmentGraph::counter = 0;

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
