#include "time_advance_table.h"

#include "segment.h"
#include "../tools/class_is_derived_from.h"

risc::sg::TimeAdvanceTable::TimeAdvanceTable(SegmentGraph &graph):
  segment_graph_(graph)
{
  size_ = boost::num_vertices(segment_graph_.graph_);

  current_time_advance_ = new Time[size_];
  next_time_advance_ = new Time[size_];

  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    table_index_segment_lookup_.push_back(vp.first);
  }

  determine_conflict_table();
}

risc::sg::TimeAdvanceTable::~TimeAdvanceTable()
{
  delete[] current_time_advance_;
  delete[] next_time_advance_;
}

void
risc::sg::TimeAdvanceTable::determine_conflict_table()
{
  // The logic how are these tables are build can be looked up in
  // "Out-of-Order Parallel Discrete Event Simulation for Transaction Level
  // Models" on page 1864

  // Here we are determining the current time advance table. The value is how
  // much time is spend to enter a new segment.
  for(int i = 0; i < size_; i++) {

    current_time_advance_[i]
      = get_waiting_time_for_segment(table_index_segment_lookup_.at(i));
  }

  // Here we are filling the next time advance table. The value will be the
  // minimum of all directly following segments
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;

    boost::tie(out_edge_begin, out_edge_end)
      = boost::out_edges(*(vp.first), segment_graph_.graph_);

    Time min_time;

    if(out_edge_begin != out_edge_end) {

      for(out_edge_iter  = out_edge_begin;
          out_edge_iter != out_edge_end;
          out_edge_iter++) {

        if(out_edge_iter == out_edge_begin) {

          // initialization for the first iteration
          min_time = get_waiting_time_for_segment(std::find(
                         vertices(segment_graph_.graph_).first,
                         vertices(segment_graph_.graph_).second,
                         boost::target(*out_edge_iter, segment_graph_.graph_)));

        } else {

          Time temp = get_waiting_time_for_segment(std::find(
                          vertices(segment_graph_.graph_).first,
                          vertices(segment_graph_.graph_).second,
                          boost::target(*out_edge_iter, segment_graph_.graph_)));

          if(temp < min_time) {
            min_time = temp;
          }
        }
      }
    } else {
      min_time = Time(true);
    }

    // we have to look up the index in the table
    int i = 0;
    for(; i<size_; i++) {

      if(*table_index_segment_lookup_.at(i) == *vp.first) {
        break;
      }
    }

    next_time_advance_[i] = min_time;
  }
}

void
risc::sg::TimeAdvanceTable::print_current_time_advance_table(
    std::string filename)
{
  print_time_table(current_time_advance_, filename);
}

void
risc::sg::TimeAdvanceTable::print_next_time_advance_table(
    std::string filename)
{
  print_time_table(next_time_advance_, filename);
}

void
risc::sg::TimeAdvanceTable::print_time_table(Time *table, std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i=0; i < table_index_segment_lookup_.size(); i++) {

    myfile << "    <th>"
           << segment_graph_.graph_[*table_index_segment_lookup_.at(i)].id_
           << "</th>" << std::endl;
  }
  myfile << "  </tr>" << std::endl;

  myfile << "  <tr>" << std::endl;
  for(int i=0; i < table_index_segment_lookup_.size(); i++) {

    myfile << "    <td>"
           << table[i]
           << "</td>" << std::endl;
  }
  myfile << "  </tr>" << std::endl;

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

risc::sg::Time
risc::sg::TimeAdvanceTable::get_waiting_time_for_segment(
    VertexIterator vertex_iter)
{
  if(!segment_graph_.graph_[*vertex_iter].boundary_node_) {
    return Time(0,0);

  }

  const SgStatement *stmt
    = isSgStatement(segment_graph_.graph_[*vertex_iter].boundary_node_);
  const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

  if(expr_stmt) {

    const SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call) {

      if(SageInterface::get_name(
            func_call->getAssociatedFunctionDeclaration()) == "wait") {

        SgExpressionPtrList &function_call_exprs
          = func_call->get_args()->get_expressions();

        // In this case we can wait for
        //   - sc_event_and_list, sc_event_or_list
        //   - sc_event
        //   - sc_zero_time
        //   - (time, time unit)
        //
        if(function_call_exprs.size() == 0) {

          assert("No support for immediate notification" && 0);

        } else if(function_call_exprs.size() == 1) {

          // SC_ZERO_TIME
          if(func_call->get_args()-> get_expressions().at(0)
              ->unparseToString() == "SC_ZERO_TIME") {

            return risc::sg::Time(0,1);

          } else {

            SgVarRefExp* waiting_event
              = isSgVarRefExp(isSgArrowExp(function_call_exprs.at(0))
                  ->get_rhs_operand());

            // SC_EVENT
            if(risc::tools::is_sc_event(waiting_event)) {

              return risc::sg::Time(0,1);
            } else {

              // TODO FIX ME
              // sc_event_and_list, sc_event_or_list
              //assert("No support for sc_event_and_list/sc_event_or_list" && 0);
            }
          }
        } else if(function_call_exprs.size() == 2) {
          // something like wait(1,SC_NS)

          SageInterface::const_int_expr_t waiting_time =
            SageInterface::evaluateConstIntegerExpression(
                isSgCastExp(function_call_exprs.at(0))->get_operand());

          if(waiting_time.hasValue_) {

            return Time(waiting_time.value_ *
                convert_si_magnitude(function_call_exprs.at(1)
                  ->unparseToString()),0);

          } else {

            assert("Wait with a non const expression" && 0);
          }

        } else {

          assert("No support case for wait time determination" && 0);
        }

        if(func_call->get_args()->get_expressions().size() == 0) {

            return Time(0,0);
        }
      }
    }
  }

  return Time(0,0);
}

int
risc::sg::TimeAdvanceTable::get_segment_id(int index)
{
  return segment_graph_.graph_[*table_index_segment_lookup_.at(index)].id_;
}

long long
risc::sg::TimeAdvanceTable::convert_si_magnitude(std::string magnitude)
{
  if("SC_FS" == magnitude) {
    return 1;

  } else if("SC_PS" == magnitude) {
    return 1000;

  } else if("SC_NS" == magnitude) {
    return 1000000;

  } else if("SC_US" == magnitude) {
    return 1000000000;

  } else if("SC_MS" == magnitude) {
    return 1000000000000;

  } else if("SC_SEC" == magnitude) {
    return 1000000000000000;

  } else {
    assert("Unsupported magnitude" && 0);
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
