#include "variable_helper.h"

//needed for DataConflictAstAttributes
#include "../segment_graph/data_conflict_ast_attributes.h"

#include "../internal_representation/design.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/primitive_channel.h"

bool
risc::tools::is_reference_variable(SgVariableSymbol *var_symbol)
{
  return isSgReferenceType(var_symbol->get_type());
}

bool
risc::tools::is_pointer_variable(SgVariableSymbol *var_symbol)
{
  return isSgPointerType(var_symbol->get_type());
}

bool
risc::tools::is_global_variable(SgVariableSymbol *var_symbol)
{
  if(SageInterface::getGlobalScope(var_symbol)->symbol_exists(var_symbol)) {

    return true;
  } else {
    return false;
  }
}

bool
risc::tools::is_local_variable(SgVariableSymbol *var_symbol)
{
  SgScopeStatement *scope =
    SageInterface::getEnclosingScope(var_symbol);

  SgBasicBlock *bb = isSgBasicBlock(scope);
  if(bb) {

    if(bb->symbol_exists(var_symbol)) {

      return true;
    } else {
      return false;
    }

  } else {
    return false;
  }
}

bool
risc::tools::is_member_variable(SgVariableSymbol *var_symbol)
{
  SgClassDefinition *class_def =
    SageInterface::getEnclosingClassDefinition(var_symbol);

  if(class_def == NULL) {
    return false;
  }

  if(class_def->symbol_exists(var_symbol)) {

    return true;
  } else {
    return false;
  }
}

bool
risc::tools::is_static_variable(SgVariableSymbol *var_symbol)
{
  SgInitializedName *init_name = var_symbol->get_declaration();

  SgVariableDefinition *var_def
    = isSgVariableDefinition(init_name->get_declptr());

  if(var_def) {
    return SageInterface::isStatic(var_def->get_declaration());
  } else {

    return init_name->get_storageModifier().isStatic();
  }
}

bool
risc::tools::is_namespace_variable(SgVariableSymbol *var_symbol)
{
  SgScopeStatement *scope
    = SageInterface::getEnclosingScope(var_symbol);

  SgNamespaceDefinitionStatement *namespace_def
    = isSgNamespaceDefinitionStatement(scope);

  if(namespace_def) {

    if(namespace_def->symbol_exists(var_symbol)) {

      return true;
    } else {
      return false;
    }

  } else {
    return false;
  }
}

bool
risc::tools::is_function_parameter_variable(SgVariableSymbol *var_symbol)
{
  SgFunctionDefinition *function_def
    = SageInterface::getEnclosingFunctionDefinition(var_symbol);

  if(function_def) {
    if(function_def->symbol_exists(var_symbol)) {

      return true;
    } else {
      return false;
    }
  }

  return false;
}

bool
risc::tools::is_member_of(SgVariableSymbol *is, SgSymbol *of)
{
  SgClassDeclaration *class_decl = NULL;

  if(isSgVariableSymbol(of)) {

    SgInitializedName *name = isSgVariableSymbol(of)->get_declaration();

    class_decl = isSgClassDeclaration(name->get_declaration());
    if(!class_decl) {
      return false;
    }

  } else if(isSgClassSymbol(of)) {
    class_decl = isSgClassSymbol(of)->get_declaration();

  } else {
    return false;
  }

  if(class_decl) {
    SgClassDefinition *class_def = class_decl->get_definition();

    if(class_def->symbol_exists(is)) {
      return true;
    } else {

      SgBaseClassPtrList &base_classes = class_def->get_inheritances();

      // iterate all direct base classes
      for(SgBaseClassPtrList::iterator
          base_class_iter  = base_classes.begin();
          base_class_iter != base_classes.end();
          base_class_iter++) {

        SgSymbol *symbol = (*base_class_iter)->get_base_class()->
                             search_for_symbol_from_symbol_table();


        if(is_member_of(is, symbol)) {
          return true;
        }
      }

      return false;
    }
  }
  return false;
}

bool risc::tools::is_loop_variable(SgVariableSymbol *var_symbol)
{
  SgInitializedName *name = var_symbol->get_declaration();
  SgDeclarationStatement *var_decl = name->get_declaration();

  SgForStatement *for_stmt = isSgForStatement(SageInterface::getEnclosingScope(var_decl));
  if(for_stmt) {

    if(var_symbol == for_stmt->lookup_variable_symbol(name->get_name())) {
      return true;
    }
  }

  return false;
}

bool risc::tools::is_stack_variable(SgVariableSymbol *var_symbol)
{
  return (tools::is_local_variable(var_symbol) && !tools::is_static_variable(var_symbol))
    || tools::is_loop_variable(var_symbol)
    || tools::is_function_parameter_variable(var_symbol);
}

bool risc::tools::is_data_segment_variable(SgVariableSymbol *var_symbol)
{
  return tools::is_global_variable(var_symbol)
    || tools::is_namespace_variable(var_symbol)
    || (tools::is_local_variable(var_symbol) && tools::is_static_variable(var_symbol))
    || (tools::is_member_variable(var_symbol) && tools::is_static_variable(var_symbol));
}

void risc::tools::analyze_conflict_free(SgVariableSymbol *symbol)
{
  risc::sg::DataConflictAstAttributes *data_conflict_attribute =
    dynamic_cast<risc::sg::DataConflictAstAttributes*>(
        symbol->getAttribute("data_conflict"));

  if(data_conflict_attribute == NULL) {

    data_conflict_attribute = new risc::sg::DataConflictAstAttributes();
    symbol->setAttribute("data_conflict", data_conflict_attribute);
  }

  SgInitializedName *name = symbol->get_declaration();
  SgDeclarationStatement *var_decl = name->get_declaration();

  SgScopeStatement *scope = SageInterface::getEnclosingScope(var_decl);
  SgStatement *previous = var_decl;

  // TODO: The function SageInterface::getPreviousStatement has a bug.
  // So, we have to do it manually.
  if(scope->containsOnlyDeclarations()) {

    for(SgDeclarationStatementPtrList::iterator
        iter  = scope->getDeclarationList().begin();
        iter != scope->getDeclarationList().end();
        iter++) {

      SgPragmaDeclaration *pragma = isSgPragmaDeclaration(*iter);

      if(pragma) {

        std::istringstream ss(pragma->get_pragma()->get_pragma());
        std::vector<std::string> unparsed(3);

        for(int i = 0; i != 3; i++) {
          ss >> unparsed.at(i);
        }

        if(unparsed.at(0) == "RISC"
            && unparsed.at(1)
                 == symbol->get_declaration()->get_name().getString()
            && unparsed.at(2) == "conflict-free") {

          data_conflict_attribute->is_conflict_free_ = true;
          return;
        }
      }
    }

  } else {
    for(SgStatementPtrList::iterator
        iter  = scope->getStatementList().begin();
        iter != scope->getStatementList().end();
        iter++) {

      SgPragmaDeclaration *pragma = isSgPragmaDeclaration(*iter);

      if(pragma) {

        std::istringstream ss(pragma->get_pragma()->get_pragma());
        std::vector<std::string> unparsed(3);

        for(int i = 0; i != 3; i++) {
          ss >> unparsed.at(i);
        }

        if(unparsed.at(0) == "RISC"
            && unparsed.at(1)
                 == symbol->get_declaration()->get_name().getString()
            && unparsed.at(2) == "conflict-free") {

          data_conflict_attribute->is_conflict_free_ = true;
          return;
        }
      }
    }
  }
}

bool risc::tools::is_channel_variable(SgVariableSymbol *symbol)
{
  SgClassDefinition *class_def
    = SageInterface::getEnclosingClassDefinition(symbol);

  if(class_def) {

    PrimitiveChannel *prim_channel
      = risc::get_design()->find_prim_channel_definition(class_def);
    HierarchicalChannel *hier_channel
      = risc::get_design()->find_channel_definition(class_def);

    if(prim_channel || hier_channel) {
      return true;
    } else {
      return false;
    }
  }
  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
